# -*- coding: utf-8 -*-
"""
batterydataextractor.nlp.tokenize

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Word and sentence tokenizers.
"""
from abc import ABCMeta, abstractmethod
import logging
import re
import six
import spacy

log = logging.getLogger(__name__)


class BaseTokenizer(six.with_metaclass(ABCMeta)):
    """Abstract base class from which all Tokenizer classes inherit.
    Subclasses must implement a ``span_tokenize(text)`` method that returns a list of integer offset tuples that
    identify tokens in the text.
    """

    def tokenize(self, s):
        """Return a list of token strings from the given sentence.
        :param string s: The sentence string to tokenize.
        :rtype: iter(str)
        """
        return [s[start:end] for start, end in self.span_tokenize(s)]

    @abstractmethod
    def span_tokenize(self, s):
        """Return a list of integer offsets that identify tokens in the given sentence.
        :param string s: The sentence string to tokenize.
        :rtype: iter(tuple(int, int))
        """
        return

    def tokenize_sents(self, strings):
        """Apply the ``tokenize`` method to each sentence in ``strings``.
        :param list(str) strings: A list of sentence strings to tokenize.
        :rtype: iter(iter(str))
        """
        return [self.tokenize(s) for s in strings]

    def span_tokenize_sents(self, strings):
        """Apply the ``span_tokenize`` method to each sentence in ``strings``.
        :param list(str) strings: A list of sentence strings to tokenize.
        :rtype: iter(iter(tuple(int, int)))
        """
        for s in strings:
            yield list(self.span_tokenize(s))


def regex_span_tokenize(s, regex):
    """Return spans that identify tokens in s split using regex."""
    left = 0
    for m in re.finditer(regex, s, re.U):
        right, next = m.span()
        if right != 0:
            yield left, right
        left = next
    yield left, len(s)


class SentenceTokenizer(BaseTokenizer):
    """Sentence tokenizer from Spacy."""

    model = 'en_core_sci_sm'  # This is available from Spacy

    def __init__(self, model=None):
        self.model = model if model is not None else self.model
        self._tokenizer = None
        log.debug('%s: Initializing with %s' % (self.__class__.__name__, self.model))

    def span_tokenize(self, s):
        """Return a list of integer offsets that identify sentences in the given text.
        :param string s: The text to tokenize into sentences.
        :rtype: iter(tuple(int, int))
        """
        if self._tokenizer is None:
            self._tokenizer = spacy.load(self.model)
        sents = list(self._tokenizer(s).sents)
        spans = [(i.start_char, i.end_char) for i in sents]
        return spans


class ChemSentenceTokenizer(SentenceTokenizer):
    """"""
    model = 'en_core_sci_sm'  # This is available from Spacy


def bracket_level(text, open={'(', '[', '{'}, close={')', ']', '}'}):
    """Return 0 if string contains balanced brackets or no brackets."""
    level = 0
    for c in text:
        if c in open:
            level += 1
        elif c in close:
            level -= 1
    return level


GREEK = {
    'Α',  # \u0391
    'Β',  # \u0392
    'Γ',  # \u0393
    'Δ',  # \u0394
    'Ε',  # \u0395
    'Ζ',  # \u0396
    'Η',  # \u0397
    'Θ',  # \u0398
    'Ι',  # \u0399
    'Κ',  # \u039a
    'Λ',  # \u039b
    'Μ',  # \u039c
    'Ν',  # \u039d
    'Ξ',  # \u039e
    'Ο',  # \u039f
    'Π',  # \u03a0
    'Ρ',  # \u03a1
    'Σ',  # \u03a3
    'Τ',  # \u03a4
    'Υ',  # \u03a5
    'Φ',  # \u03a6
    'Χ',  # \u03a7
    'Ψ',  # \u03a8
    'Ω',  # \u03a9
    'α',  # \u03b1
    'β',  # \u03b2
    'γ',  # \u03b3
    'δ',  # \u03b4
    'ε',  # \u03b5
    'ζ',  # \u03b6
    'η',  # \u03b7
    'θ',  # \u03b8
    'ι',  # \u03b9
    'κ',  # \u03ba
    'λ',  # \u03bb
    'μ',  # \u03bc
    'ν',  # \u03bd
    'ξ',  # \u03be
    'ο',  # \u03bf
    'π',  # \u03c0
    'ρ',  # \u03c1
    'σ',  # \u03c3
    'τ',  # \u03c4
    'υ',  # \u03c5
    'φ',  # \u03c6
    'χ',  # \u03c7
    'ψ',  # \u03c8
    'ω',  # \u03c9
}


class WordTokenizer(BaseTokenizer):
    """Standard word tokenizer for generic English text."""
    #: Split before and after these sequences, wherever they occur, unless entire token is one of these sequences
    SPLIT = [
        '----',
        '––––',  # \u2013 en dash
        '————',  # \u2014 em dash
        '<--->',
        '---',
        '–––',  # \u2013 en dash
        '———',  # \u2014 em dash
        '<-->',
        '-->',
        '...',
        '--',
        '––',  # \u2013 en dash
        '——',  # \u2014 em dash
        '``',
        "''",
        '->',
        '<',
        '>',
        '–',  # \u2013 en dash
        '—',  # \u2014 em dash
        '―',  # \u2015 horizontal bar
        '~',  # \u007e Tilde
        '⁓',  # \u2053 Swung dash
        '∼',  # \u223c Tilde operator
        '°',  # \u00b0 Degrees
        ';',
        '@',
        '#',
        '$',
        '£',  # \u00a3
        '€',  # \u20ac
        '%',
        '&',
        '?',
        '!',
        '™',  # \u2122
        '®',  # \u00ae
        '…',  # \u2026
        '⋯',  # \u22ef Mid-line ellipsis
        '†',  # \u2020 Dagger
        '‡',  # \u2021 Double dagger
        '§',  # \u00a7 Section sign
        '¶'   # \u00b6 Pilcrow sign
        '≠',  # \u2260
        '≡',  # \u2261
        '≢',  # \u2262
        '≣',  # \u2263
        '≤',  # \u2264
        '≥',  # \u2265
        '≦',  # \u2266
        '≧',  # \u2267
        '≨',  # \u2268
        '≩',  # \u2269
        '≪',  # \u226a
        '≫',  # \u226b
        '≈',  # \u2248
        '=',
        '÷',  # \u00f7
        '×',  # \u00d7
        '→',  # \u2192
        '⇄',  # \u21c4
        '"',  # \u0022 Quote mark
        '“',  # \u201c
        '”',  # \u201d
        '„',  # \u201e
        '‟',  # \u201f
        '‘',  # \u2018 Left single quote
        # '’',  # \u2019 Right single quote  - Regularly used as an apostrophe, so don't always split
        '‚',  # \u201a Single low quote
        '‛',  # \u201b Single reversed quote
        '`',  # \u0060
        '´',  # \u00b4
        # Primes
        '′',  # \u2032
        '″',  # \u2033
        '‴',  # \u2034
        '‵',  # \u2035
        '‶',  # \u2036
        '‷',  # \u2037
        '⁗',  # \u2057
        # Brackets
        '(',
        '[',
        '{',
        '}',
        ']',
        ')',
        # Slashes
        '/',  # \u002f Solidus
        '⁄',  # \u2044 Fraction slash
        '∕',  # \u2215 Division slash
        # Hyphens and Minuses
        # '-',  # \u002d Hyphen-minus
        '−',  # \u2212 Minus
        '‒',  # \u2012 figure dash
        # '‐',  # \u2010 Hyphen
        # '‐',  # \u2010 Hyphen
        # '‑',  # \u2011 Non-breaking hyphen
        '+',  # \u002b Plus
        '±',  # \u00b1 Plus/Minus
    ]
    #: Split around these sequences unless they are followed by a digit
    SPLIT_NO_DIGIT = [':', ',']
    #: Split after these sequences if they start a word
    SPLIT_START_WORD = ["''", "``", "'"]
    #: Split before these sequences if they end a word
    SPLIT_END_WORD = ["'s", "'m", "'d", "'ll", "'re", "'ve", "n't", "''", "'", "’s", "’m", "’d", "’ll", "’re", "’ve", "n’t", "’", "’’"]
    #: Don't split full stop off last token if it is one of these sequences
    NO_SPLIT_STOP = ['...', 'al.', 'Co.', 'Ltd.', 'Pvt.', 'A.D.', 'B.C.', 'B.V.', 'S.D.', 'U.K.', 'U.S.', 'r.t.']
    #: Split these contractions at the specified index
    CONTRACTIONS = [("cannot", 3), ("d'ye", 1), ("d’ye", 1), ("gimme", 3), ("gonna", 3), ("gotta", 3), ("lemme", 3),
                    ("mor'n", 3), ("mor’n", 3), ("wanna", 3), ("'tis", 2), ("'twas", 2)]
    #: Don't split these sequences.
    NO_SPLIT = {
        'mm-hm', 'mm-mm', 'o-kay', 'uh-huh', 'uh-oh', 'wanna-be'
    }
    #: Don't split around hyphens with these prefixes
    NO_SPLIT_PREFIX = {
        'e', 'a', 'u', 'x', 'agro', 'ante', 'anti', 'arch', 'be', 'bi', 'bio', 'co', 'counter', 'cross', 'cyber',
        'de', 'eco', 'ex', 'extra', 'inter', 'intra', 'macro', 'mega', 'micro', 'mid', 'mini', 'multi', 'neo', 'non',
        'over', 'pan', 'para', 'peri', 'post', 'pre', 'pro', 'pseudo', 'quasi', 're', 'semi', 'sub', 'super', 'tri',
        'ultra', 'un', 'uni', 'vice'
    }
    #: Don't split around hyphens with these suffixes.
    NO_SPLIT_SUFFIX = {
        'esque', 'ette', 'fest', 'fold', 'gate', 'itis', 'less', 'most', '-o-torium', 'rama', 'wise'
    }
    #: Don't split around hyphens if only these characters before or after.
    NO_SPLIT_CHARS = '0123456789,\'"“”„‟‘’‚‛`´′″‴‵‶‷⁗'

    def __init__(self, split_last_stop=True):
        #: Whether to split off the final full stop (unless preceded by NO_SPLIT_STOP). Default True.
        self.split_last_stop = split_last_stop

    def _split_span(self, span, index, length=0):
        """Split a span into two or three separate spans at certain indices."""
        offset = span[1] + index if index < 0 else span[0] + index
        # log.debug([(span[0], offset), (offset, offset + length), (offset + length, span[1])])
        return [(span[0], offset), (offset, offset + length), (offset + length, span[1])]

    def _subspan(self, s, span, nextspan):
        """Recursively subdivide spans based on a series of rules."""
        text = s[span[0]:span[1]]
        lowertext = text.lower()

        # Skip if only a single character or a split sequence
        if span[1] - span[0] < 2 or text in self.SPLIT or text in self.SPLIT_END_WORD or text in self.SPLIT_START_WORD or lowertext in self.NO_SPLIT:
            return [span]

        # Skip if it looks like URL
        if text.startswith('http://') or text.startswith('ftp://') or text.startswith('www.'):
            return [span]

        # Split full stop at end of final token (allow certain characters to follow) unless ellipsis
        if self.split_last_stop and nextspan is None and text not in self.NO_SPLIT_STOP and not text[-3:] == '...':
            if text[-1] == '.':
                return self._split_span(span, -1)
            ind = text.rfind('.')
            if ind > -1 and all(t in '\'‘’"“”)]}' for t in text[ind + 1:]):
                return self._split_span(span, ind, 1)

        # Split off certain sequences at the end of a word
        for spl in self.SPLIT_END_WORD:
            if text.endswith(spl) and len(text) > len(spl) and text[-len(spl) - 1].isalpha():
                return self._split_span(span, -len(spl), 0)

        # Split off certain sequences at the start of a word
        for spl in self.SPLIT_START_WORD:
            if text.startswith(spl) and len(text) > len(spl) and text[-len(spl) - 1].isalpha():
                return self._split_span(span, len(spl), 0)

        # Split around certain sequences
        for spl in self.SPLIT:
            ind = text.find(spl)
            if ind > -1:
                return self._split_span(span, ind, len(spl))

        # Split around certain sequences unless followed by a digit
        for spl in self.SPLIT_NO_DIGIT:
            ind = text.rfind(spl)
            if ind > -1 and (len(text) <= ind + len(spl) or not text[ind + len(spl)].isdigit()):
                return self._split_span(span, ind, len(spl))

        # Characters to split around, but with exceptions
        for i, char in enumerate(text):
            if char == '-':
                before = lowertext[:i]
                after = lowertext[i+1:]
                # By default we split on hyphens
                split = True
                if before in self.NO_SPLIT_PREFIX or after in self.NO_SPLIT_SUFFIX:
                    split = False  # Don't split if prefix or suffix in list
                elif not before.strip(self.NO_SPLIT_CHARS) or not after.strip(self.NO_SPLIT_CHARS):
                    split = False  # Don't split if prefix or suffix entirely consist of certain characters
                if split:
                    return self._split_span(span, i, 1)

        # Split contraction words
        for contraction in self.CONTRACTIONS:
            if lowertext == contraction[0]:
                return self._split_span(span, contraction[1])
        return [span]

    def span_tokenize(self, s):
        """"""
        # First get spans by splitting on all whitespace
        # Includes: \u0020 \u00A0 \u1680 \u180E \u2000 \u2001 \u2002 \u2003 \u2004 \u2005 \u2006 \u2007 \u2008 \u2009 \u200A \u202F \u205F \u3000
        spans = [(left, right) for left, right in regex_span_tokenize(s, '\s+') if not left == right]
        i = 0
        # Recursively split spans according to rules
        while i < len(spans):
            subspans = self._subspan(s, spans[i], spans[i + 1] if i + 1 < len(spans) else None)
            spans[i:i+1] = [subspan for subspan in subspans if subspan[1] - subspan[0] > 0]
            if len(subspans) == 1:
                i += 1
        return spans


class ChemWordTokenizer(WordTokenizer):
    """Word Tokenizer for chemistry text."""

    #: Split before and after these sequences, wherever they occur, unless entire token is one of these sequences
    SPLIT = [
        '----',
        '––––',  # \u2013 en dash
        '————',  # \u2014 em dash
        '<--->',
        '---',
        '–––',  # \u2013 en dash
        '———',  # \u2014 em dash
        '<-->',
        '-->',
        '...',
        '--',
        '––',  # \u2013 en dash
        '——',  # \u2014 em dash
        # '``',
        # "''",
        # '->',  # Don't split around this if occurs within chemical name
        '<',
        # '>',   # Don't split around this if occurs within chemical name
        ').',  # Fix missing whitespace errors
        '.(',  # Fix missing whitespace errors
        '–',  # \u2013 en dash
        '—',  # \u2014 em dash
        '―',  # \u2015 horizontal bar
        '~',  # \u007e Tilde
        '⁓',  # \u2053 Swung dash
        '∼',  # \u223c Tilde operator
        '°',  # \u00b0 Degrees
        # ';',
        '@',
        '#',
        '$',
        '£',  # \u00a3
        '€',  # \u20ac
        '%',
        '&',
        '?',
        '!',
        '™',  # \u2122
        '®',  # \u00ae
        '…',  # \u2026
        '⋯',  # \u22ef Mid-line ellipsis
        '†',  # \u2020 Dagger
        '‡',  # \u2021 Double dagger
        '§',  # \u00a7 Section sign
        '¶'   # \u00b6 Pilcrow sign
        '≠',  # \u2260
        '≡',  # \u2261
        '≢',  # \u2262
        '≣',  # \u2263
        '≤',  # \u2264
        '≥',  # \u2265
        '≦',  # \u2266
        '≧',  # \u2267
        '≨',  # \u2268
        '≩',  # \u2269
        '≪',  # \u226a
        '≫',  # \u226b
        '≈',  # \u2248
        '=',
        '÷',  # \u00f7
        '×',  # \u00d7
        # '→',  # \u2192 # Don't split around this if occurs within chemical name
        '⇄',  # \u21c4
        '"',  # \u0022 Quote mark
        '“',  # \u201c
        '”',  # \u201d
        '„',  # \u201e
        '‟',  # \u201f
        '‘',  # \u2018 Left single quote
        # '’',  # \u2019 Right single quote - Regularly used as an apostrophe, so don't always split
        '‚',  # \u201a Single low quote
        '‛',  # \u201b Single reversed quote
        '`',  # \u0060
        '´',  # \u00b4
        # Primes
        # '′',  # \u2032
        # '″',  # \u2033
        # '‴',  # \u2034
        # '‵',  # \u2035
        # '‶',  # \u2036
        # '‷',  # \u2037
        # '⁗',  # \u2057
        # Brackets
        # '(',
        # '[',
        # '{',
        # '}',
        # ']',
        # ')',
        # Slashes
        # '/',  # \u002f Solidus
        # '⁄',  # \u2044 Fraction slash
        # '∕',  # \u2215 Division slash
        # Hyphens and Minuses
        # '-',  # \u002d Hyphen-minus
        # '−',  # \u2212 Minus
        # '‒',  # \u2012 figure dash
        # '‐',  # \u2010 Hyphen
        # '‑',  # \u2011 Non-breaking hyphen
        # '+',  # \u002b Plus
        # '±',  # \u00b1 Plus/Minus
    ]
    #: Split before these sequences if they end a token
    SPLIT_END = [':', ',', '(TM)', '(R)', '(®)', '(™)', '(■)', '(◼)', '(●)', '(▲)', '(○)', '(◆)', '(▼)', '(⧫)', '(△)', '(◇)', '(▽)', '(⬚)', '(×)', '(□)', '(•)', '’']
    #: Split before these sequences if they end a token, unless preceded by a digit
    SPLIT_END_NO_DIGIT = ['(aq)', '(aq.)', '(s)', '(l)', '(g)']
    #: Don't split around slash when both preceded and followed by these characters
    NO_SPLIT_SLASH = ['+', '-', '−']
    #: Regular expression that matches a numeric quantity with units
    QUANTITY_RE = re.compile(r'^((\d\d\d)g|([-−]?\d+\.\d+|10[-−]\d+)(g|s|m|N|V)([-−]?[1-4])?|(\d*[-−]?\d+\.?\d*)([pnµμm]A|[µμmk]g|[kM]J|m[lL]|[nµμm]?M|[nµμmc]m|kN|[mk]V|[mkMG]?W|[mnpμµ]s|Hz|[Mm][Oo][Ll](e|ar)?s?|k?Pa|ppm|min)([-−]?[1-4])?)$')
    #: Don't split on hyphen if the prefix matches this regular expression
    NO_SPLIT_PREFIX_ENDING = re.compile('(^\(.*\)|^[\d,\'"“”„‟‘’‚‛`´′″‴‵‶‷⁗Α-Ωα-ω]+|ano|ato|azo|boc|bromo|cbz|chloro|eno|fluoro|fmoc|ido|ino|io|iodo|mercapto|nitro|ono|oso|oxalo|oxo|oxy|phospho|telluro|tms|yl|ylen|ylene|yliden|ylidene|ylidyn|ylidyne)$', re.U)
    #: Don't split on hyphen if prefix or suffix match this regular expression
    NO_SPLIT_CHEM = re.compile('([\-α-ω]|\d+,\d+|\d+[A-Z]|^d\d\d?$|acetic|acetyl|acid|acyl|anol|azo|benz|bromo|carb|cbz|chlor|cyclo|ethan|ethyl|fluoro|fmoc|gluc|hydro|idyl|indol|iene|ione|iodo|mercapto|n,n|nitro|noic|o,o|oxalo|oxo|oxy|oyl|onyl|phen|phth|phospho|pyrid|telluro|tetra|tms|ylen|yli|zole|alpha|beta|gamma|delta|epsilon|theta|kappa|lambda|sigma|omega)', re.U | re.I)
    #: Don't split on hyphen if the prefix is one of these sequences
    NO_SPLIT_PREFIX = {
        'e', 'a', 'u', 'x', 'agro', 'ante', 'anti', 'arch', 'be', 'bi', 'bio', 'co', 'counter', 'cross', 'cyber',
        'de', 'eco', 'ex', 'extra', 'inter', 'intra', 'macro', 'mega', 'micro', 'mid', 'mini', 'multi', 'neo', 'non',
        'over', 'pan', 'para', 'peri', 'post', 'pre', 'pro', 'pseudo', 'quasi', 're', 'semi', 'sub', 'super', 'tri',
        'ultra', 'un', 'uni', 'vice',

        'aci', 'adeno', 'aldehydo', 'allo', 'alpha', 'altro', 'ambi', 'aorto', 'arachno', 'as', 'beta', 'bis', 'catena',
        'centi', 'chi', 'chiro', 'circum', 'cis', 'closo', 'colo', 'conjuncto', 'conta', 'contra', 'cortico', 'cosa',
        'counter', 'cran', 'crypto', 'cyclo', 'deca', 'deci', 'delta', 'demi', 'di', 'dis', 'dl', 'eco', 'electro',
        'endo', 'ennea', 'ent', 'epi', 'epsilon', 'erythro', 'eta', 'exo', 'ferro', 'galacto', 'gamma', 'gastro',
        'giga', 'gluco', 'glycero', 'graft', 'gulo', 'hemi', 'hepta', 'hexa', 'homo', 'hydro', 'hypho', 'hypo', 'ideo',
        'idio', 'in', 'infra', 'iota', 'iso', 'judeo', 'kappa', 'keto', 'kis', 'lambda', 'lyxo', 'manno', 'medi',
        'meso', 'meta', 'milli', 'mono', 'mu', 'muco', 'musculo', 'myo', 'nano', 'neuro', 'nido', 'nitro', 'nona',
        'nor', 'novem', 'novi', 'nu', 'octa', 'octi', 'octo', 'omega', 'omicron', 'ortho', 'paleo', 'pelvi', 'penta',
        'pheno', 'phi', 'pi', 'pica', 'pneumo', 'poly', 'preter', 'psi', 'quadri', 'quater', 'quinque', 'recto', 'rho',
        'ribo', 'salpingo', 'scyllo', 'sec', 'sept', 'septi', 'sero', 'sesqui', 'sexi', 'sigma', 'sn', 'soci', 'supra',
        'sur', 'sym', 'syn', 'talo', 'tau', 'tele', 'ter', 'tera', 'tert', 'tetra', 'theta', 'threo', 'trans',
        'triangulo', 'tris', 'uber', 'unsym', 'upsilon', 'veno', 'ventriculo', 'xi', 'xylo', 'zeta',
    }
    #: Split on hyphens followed by one of these sequences
    SPLIT_SUFFIX = {
        'absorption', 'abstinent', 'abstraction', 'abuse', 'accelerated', 'accepting', 'acclimated', 'acclimation',
        'acid', 'activated', 'activation', 'active', 'activity', 'addition', 'adducted', 'adducts', 'adequate',
        'adjusted', 'administrated', 'adsorption', 'affected', 'aged', 'alcohol', 'alcoholic', 'algae', 'alginate',
        'alkaline', 'alkylated', 'alkylation', 'alkyne', 'analogous', 'anesthetized', 'appended', 'armed', 'aromatic',
        'assay', 'assemblages', 'assisted', 'associated', 'atom', 'atoms', 'attenuated', 'attributed', 'backbone',
        'base', 'based', 'bearing', 'benzylation', 'binding', 'biomolecule', 'biotic', 'blocking', 'blood', 'bond',
        'bonded', 'bonding', 'bonds', 'boosted', 'bottle', 'bottled', 'bound', 'bridge', 'bridged', 'buffer',
        'buffered', 'caged', 'cane', 'capped', 'capturing', 'carrier', 'carrying', 'catalysed', 'catalyzed', 'cation',
        'caused', 'centered', 'challenged', 'chelating', 'cleaving', 'coated', 'coating', 'coenzyme', 'competing',
        'competitive', 'complex', 'complexes', 'compound', 'compounds', 'concentration', 'conditioned', 'conditions',
        'conducting', 'configuration', 'confirmed', 'conjugate', 'conjugated', 'conjugates', 'connectivity',
        'consuming', 'contained', 'containing', 'contaminated', 'control', 'converting', 'coordinate', 'coordinated',
        'copolymer', 'copolymers', 'core', 'cored', 'cotransport', 'coupled', 'covered', 'crosslinked', 'cyclized',
        'damaged', 'dealkylation', 'decocted', 'decorated', 'deethylation', 'deficiency', 'deficient', 'defined',
        'degrading', 'demethylated', 'demethylation', 'dendrimer', 'density', 'dependant', 'dependence', 'dependent',
        'deplete', 'depleted', 'depleting', 'depletion', 'depolarization', 'depolarized', 'deprived', 'derivatised',
        'derivative', 'derivatives', 'derivatized', 'derived', 'desorption', 'detected', 'devalued', 'dextran',
        'dextrans', 'diabetic', 'dimensional', 'dimer', 'distribution', 'divalent', 'domain', 'dominated',
        'donating', 'donor', 'dopant', 'doped', 'doping', 'dosed', 'dot', 'drinking', 'driven', 'drug', 'drugs', 'dye',
        'edge', 'efficiency', 'electrodeposited', 'electrolyte', 'elevating', 'elicited', 'embedded', 'emersion',
        'emitting', 'encapsulated', 'encapsulating', 'enclosed', 'enhanced', 'enhancing', 'enriched', 'enrichment',
        'enzyme', 'epidermal', 'equivalents', 'etched', 'ethanolamine', 'evoked', 'exchange', 'excimer', 'excluder',
        'expanded', 'experimental', 'exposed', 'exposure', 'expressing', 'extract', 'extraction', 'fed', 'finger', 'fixed', 'fixing',
        'flanking', 'flavonoid', 'fluorescence', 'formation', 'forming', 'fortified', 'free', 'function',
        'functionalised', 'functionalized', 'functionalyzed', 'fused', 'gas', 'gated', 'generating', 'glucuronidating',
        'glycoprotein', 'glycosylated', 'glycosylation', 'gradient', 'grafted', 'group', 'groups', 'halogen',
        'heterocyclic', 'homologues', 'hydrogel', 'hydrolyzing', 'hydroxylated', 'hydroxylation', 'hydroxysteroid',
        'immersion', 'immobilized', 'immunoproteins', 'impregnated', 'imprinted', 'inactivated', 'increased',
        'increasing', 'incubated', 'independent', 'induce', 'induced', 'inducible', 'inducing', 'induction', 'influx',
        'inhibited', 'inhibitor', 'inhibitory', 'initiated', 'injected', 'insensitive', 'insulin', 'integrated',
        'interlinked', 'intermediate', 'intolerant', 'intoxicated', 'ion', 'ions', 'island', 'isomer', 'isomers',
        'knot', 'label', 'labeled', 'labeling', 'labelled', 'laden', 'lamp', 'laser', 'layer', 'layers', 'lesioned',
        'ligand', 'ligated', 'like', 'limitation', 'limited', 'limiting', 'lined', 'linked', 'linker', 'lipid',
        'lipids', 'lipoprotein', 'liposomal', 'liposomes', 'liquid', 'liver', 'loaded', 'loading', 'locked', 'loss',
        'lowering', 'lubricants', 'luminance', 'luminescence', 'maintained', 'majority', 'making', 'mannosylated',
        'material', 'mediated', 'metabolizing', 'metal', 'metallized', 'methylation', 'migrated', 'mimetic',
        'mimicking', 'mixed', 'mixture', 'mode', 'model', 'modified', 'modifying', 'modulated', 'moiety', 'molecule',
        'monoadducts', 'monomer', 'mutated', 'nanogel', 'nanoparticle', 'nanotube', 'need', 'negative', 'nitrosated',
        'nitrosation', 'nitrosylation', 'nmr', 'noncompetitive', 'normalized', 'nuclear', 'nucleoside', 'nucleosides',
        'nucleotide', 'nucleotides', 'nutrition', 'olefin', 'olefins', 'oligomers', 'omitted', 'only',
        'outcome', 'overload', 'oxidation', 'oxidized', 'oxo-mediated', 'oxygenation', 'page', 'paired', 'pathway',
        'patterned', 'peptide', 'permeabilized', 'permeable', 'phase', 'phospholipids', 'phosphopeptide',
        'phosphorylated', 'pillared', 'placebo', 'planted', 'plasma', 'polymer', 'polymers', 'poor', 'porous',
        'position', 'positive', 'postlabeling', 'precipitated', 'preferring', 'pretreated', 'primed', 'produced',
        'producing', 'production', 'promoted', 'promoting', 'protected', 'protein', 'proteomic', 'protonated',
        'provoked', 'purified', 'radical', 'reacting', 'reaction', 'reactive', 'reagents', 'rearranged', 'receptor',
        'receptors', 'recognition', 'redistribution', 'redox', 'reduced', 'reducing', 'reduction', 'refractory',
        'refreshed', 'regenerating', 'regulated', 'regulating', 'regulatory', 'related', 'release', 'releasing',
        'replete', 'requiring', 'resistance', 'resistant', 'resitant', 'response', 'responsive', 'responsiveness',
        'restricted', 'resulted', 'retinal', 'reversible', 'ribosylated', 'ribosylating', 'ribosylation', 'rich',
        'right', 'ring', 'saturated', 'scanning', 'scavengers', 'scavenging', 'sealed', 'secreting', 'secretion',
        'seeking', 'selective', 'selectivity', 'semiconductor', 'sensing', 'sensitive', 'sensitized', 'soluble',
        'solution', 'solvent', 'sparing', 'specific', 'spiked', 'stabilised', 'stabilized', 'stabilizing', 'stable',
        'stained', 'steroidal', 'stimulated', 'stimulating', 'storage', 'stressed', 'stripped', 'substituent',
        'substituted', 'substitution', 'substrate', 'sufficient', 'sugar', 'sugars', 'supplemented', 'supported',
        'suppressed', 'surface', 'susceptible', 'sweetened', 'synthesizing', 'tagged', 'target', 'telopeptide',
        'terminal', 'terminally', 'terminated', 'termini', 'terminus', 'ternary', 'terpolymer', 'tertiary', 'tested',
        'testes', 'tethered', 'tetrabrominated', 'tolerance', 'tolerant', 'toxicity', 'toxin', 'tracer', 'transfected',
        'transfer', 'transition', 'transport', 'transporter', 'treated', 'treating', 'treatment', 'triggered',
        'turn', 'type', 'unesterified', 'untreated', 'vacancies', 'vacancy', 'variable', 'water', 'yeast', 'yield',
        'zwitterion'
    }

    def _closing_bracket_index(self, text, bpair=('(', ')')):
        """Return the index of the closing bracket that matches the opening bracket at the start of the text."""
        level = 1
        for i, char in enumerate(text[1:]):
            if char == bpair[0]:
                level += 1
            elif char == bpair[1]:
                level -= 1
            if level == 0:
                return i + 1

    def _opening_bracket_index(self, text, bpair=('(', ')')):
        """Return the index of the opening bracket that matches the closing bracket at the end of the text."""
        level = 1
        for i, char in enumerate(reversed(text[:-1])):
            if char == bpair[1]:
                level += 1
            elif char == bpair[0]:
                level -= 1
            if level == 0:
                return len(text) - i - 2

    def _is_number(self, text):
        """Return True if the text is a number."""
        try:
            float(text)
            return True
        except ValueError:
            return False

    def _is_saccharide_arrow(self, before, after):
        """Return True if the arrow is in a chemical name."""
        if (before and after and before[-1].isdigit() and after[0].isdigit() and
            before.rstrip('0123456789').endswith('(') and after.lstrip('0123456789').startswith(')-')):
            return True
        else:
            return False

    def _subspan(self, s, span, nextspan):
        """Recursively subdivide spans based on a series of rules."""
        text = s[span[0]:span[1]]
        lowertext = text.lower()

        # Skip if only a single character or a split sequence
        if span[1] - span[0] < 2 or text in self.SPLIT or text in self.SPLIT_END_WORD or text in self.SPLIT_START_WORD or lowertext in self.NO_SPLIT:
            return [span]

        # Skip if it looks like URL
        if text.startswith('http://') or text.startswith('ftp://') or text.startswith('www.'):
            return [span]

        # Split full stop at end of final token (allow certain characters to follow) unless ellipsis
        if self.split_last_stop and nextspan is None and text not in self.NO_SPLIT_STOP and not text[-3:] == '...':
            if text[-1] == '.':
                return self._split_span(span, -1)
            ind = text.rfind('.')
            if ind > -1 and all(t in '\'‘’"“”)]}' for t in text[ind + 1:]):
                return self._split_span(span, ind, 1)

        # Split off certain sequences at the end of a token
        for spl in self.SPLIT_END:
            if text.endswith(spl) and len(text) > len(spl):
                return self._split_span(span, -len(spl), 0)

        # Split off certain sequences at the end of a word
        for spl in self.SPLIT_END_WORD:
            if text.endswith(spl) and len(text) > len(spl) and text[-len(spl) - 1].isalpha():
                return self._split_span(span, -len(spl), 0)

        # Split off certain sequences at the end of a word
        for spl in self.SPLIT_START_WORD:
            if text.startswith(spl) and len(text) > len(spl) and text[-len(spl) - 1].isalpha():
                return self._split_span(span, len(spl), 0)

        # Split around certain sequences
        for spl in self.SPLIT:
            ind = text.find(spl)
            if ind > -1:
                return self._split_span(span, ind, len(spl))

        # Split around certain sequences unless followed by a digit
        # - We skip this because of difficulty with chemical names.
        # for spl in self.SPLIT_NO_DIGIT:
        #     ind = text.rfind(spl)
        #     if ind > -1 and (len(text) <= ind + len(spl) or not text[ind + len(spl)].isdigit()):
        #         return self._split_span(span, ind, len(spl))

        # Split off certain sequences at the end of a token unless preceded by a digit
        for spl in self.SPLIT_END_NO_DIGIT:
            if text.endswith(spl) and len(text) > len(spl) and not text[-len(spl) - 1].isdigit():
                return self._split_span(span, -len(spl), 0)

        # Regular Bracket at both start and end, break off both provided they correspond
        if text.startswith('(') and text.endswith(')') and self._closing_bracket_index(text) == len(text) - 1:
            return self._split_span(span, 1, len(text)-2)

        # Split things like IR(KBr)
        if text.startswith('IR(') and text.endswith(')'):
            return self._split_span(span, 2, 1)

        # Split things like \d+\.\d+([a-z]+) e.g. UV-vis/IR peaks with bracketed strength/shape
        m = re.match('^(\d+\.\d+|\d{1,})(\([a-z]+\)|c)$', text, re.I)
        if m:
            return self._split_span(span, m.start(2), 1)

        # Split brackets off start and end if the corresponding bracket isn't within token
        for bpair in [('(', ')'), ('{', '}'), ('[', ']')]:
            #level = bracket_level(text, open=[bpair[0]], close=[bpair[1]])
            # Bracket at start, bracketlevel > 0, break it off
            if text.startswith(bpair[0]) and self._closing_bracket_index(text, bpair=bpair) is None:
                return self._split_span(span, 1, 0)
            # Bracket at end, bracketlevel < 0, break it off
            if text.endswith(bpair[1]) and self._opening_bracket_index(text, bpair=bpair) is None:
                return self._split_span(span, -1, 0)

        # Characters to split around, but with exceptions
        for i, char in enumerate(text):
            before = text[:i]
            after = text[i+1:]
            if char in {':', ';'}:
                # Split around colon unless it looks like we're in a chemical name
                if not (before and after and after[0].isdigit() and before.rstrip('′\'')[-1:].isdigit() and '-' in after) and not (self.NO_SPLIT_CHEM.search(before) and self.NO_SPLIT_CHEM.search(after)):
                    return self._split_span(span, i, 1)
            elif char in {'x', '+', '−'}:
                # Split around x, +, − (\u2212 minus) between two numbers or at start followed by numbers
                if (i == 0 or self._is_number(before)) and self._is_number(after):
                    return self._split_span(span, i, 1)
                # Also plit around − (\u2212 minus) between two letters
                if char == '−' and before and before[-1].isalpha() and after and after[0].isalpha():
                    return self._split_span(span, i, 1)
            elif char == '±':
                # Split around ± unless surrounded by brackets
                if not (before and after and before[-1] == '(' and after[0] == ')'):
                    return self._split_span(span, i, 1)
            elif char == '/':
                # Split around / unless '+/-' or '-/-' etc.
                if not (before and after and before[-1] in self.NO_SPLIT_SLASH and after[0] in self.NO_SPLIT_SLASH):
                    return self._split_span(span, i, 1)
            elif char == '>':
                if not (before and before[-1] == '-'):
                    # Split if preceding is not -
                    return self._split_span(span, i, 1)
                if before and before[-1] == '-':
                    # If preceding is -, split around -> unless in chemical name
                    if not text == '->' and not self._is_saccharide_arrow(before[:-1], after):
                        return self._split_span(span, i-1, 2)
            elif char == '→' and not self._is_saccharide_arrow(before, after):
                # Split around → unless in chemical name
                return self._split_span(span, i, 1)
            elif char == '(' and self._is_number(before) and not '(' in after and not ')' in after:
                # Split around open bracket after a number
                return self._split_span(span, i, 1)
            elif char == '-':
                lowerbefore = lowertext[:i]
                lowerafter = lowertext[i+1:]
                # Always split on -of-the- -to- -in- -by- -of- -or- -and- -per- -the-
                if lowerafter[:7] == 'of-the-':
                    return [(span[0], span[0] + i), (span[0] + i, span[0] + i + 1), (span[0] + i + 1, span[0] + i + 3), (span[0] + i + 3, span[0] + i + 4), (span[0] + i + 4, span[0] + i + 7), (span[0] + i + 7, span[0] + i + 8), (span[0] + i + 8, span[1])]
                if lowerafter[:5] in {'on-a-', 'of-a-'}:
                    return [(span[0], span[0] + i), (span[0] + i, span[0] + i + 1), (span[0] + i + 1, span[0] + i + 3), (span[0] + i + 3, span[0] + i + 4), (span[0] + i + 4, span[0] + i + 5), (span[0] + i + 5, span[0] + i + 6), (span[0] + i + 6, span[1])]
                if lowerafter[:3] in {'to-', 'in-', 'by-', 'of-', 'or-', 'on-'}:
                    return [(span[0], span[0] + i), (span[0] + i, span[0] + i + 1), (span[0] + i + 1, span[0] + i + 3), (span[0] + i + 3, span[0] + i + 4), (span[0] + i + 4, span[1])]
                if lowerafter[:4] in {'and-', 'per-', 'the-'}:
                    return [(span[0], span[0] + i), (span[0] + i, span[0] + i + 1), (span[0] + i + 1, span[0] + i + 4), (span[0] + i + 4, span[0] + i + 5), (span[0] + i + 5, span[1])]

                # By default we split on hyphens
                split = True
                if lowerafter == 'nmr':
                    split = True  # Always split NMR off end
                elif bracket_level(text) == 0 and (not bracket_level(after) == 0 or not bracket_level(before) == 0):
                    split = False  # Don't split if within brackets
                elif after and after[0] == '>':
                    split = False  # Don't split if followed by >
                elif lowerbefore in self.NO_SPLIT_PREFIX or lowerafter in self.NO_SPLIT_SUFFIX:
                    split = False  # Don't split if prefix or suffix in list
                elif self.NO_SPLIT_PREFIX_ENDING.search(lowerbefore):
                    split = False  # Don't split if prefix ends with pattern
                elif lowerafter in self.SPLIT_SUFFIX:
                    split = True  # Do split if suffix in list
                elif len(before) <= 1 or len(after) <= 2:
                    split = False  # Don't split if not at least 2 char before and 3 after
                elif self.NO_SPLIT_CHEM.search(lowerbefore) or self.NO_SPLIT_CHEM.search(lowerafter):
                    split = False  # Don't split if prefix or suffix match chem regex
                if split:
                    return self._split_span(span, i, 1)

        # Split units off the end of a numeric value
        quantity = self.QUANTITY_RE.search(text)
        if quantity:
            return self._split_span(span, len(quantity.group(6) or quantity.group(3) or quantity.group(2)), 0)

        # Split pH off the start of a numeric value
        if text.startswith('pH') and self._is_number(text[2:]):
            return self._split_span(span, 2, 0)

        # Split contraction words
        for contraction in self.CONTRACTIONS:
            if lowertext == contraction[0]:
                return self._split_span(span, contraction[1])

        if nextspan:
            nexttext = s[nextspan[0]:nextspan[1]]
            # Split NMR isotope whitespace errors (joined with previous sentence full stop)
            if nexttext == 'NMR':
                ind = text.rfind('.')
                if ind > -1 and text[ind + 1:] in {'1H', '13C', '15N', '31P', '19F', '11B', '29Si', '170', '73Ge', '195Pt', '33S', '13C{1H}'}:
                    return self._split_span(span, ind, 1)


        return [span]


class FineWordTokenizer(WordTokenizer):
    """Word Tokenizer that also split around hyphens and all colons."""
    #: Split before and after these sequences, wherever they occur, unless entire token is one of these sequences
    SPLIT = [
        '----',
        '––––',  # \u2013 en dash
        '————',  # \u2014 em dash
        '<--->',
        '---',
        '–––',  # \u2013 en dash
        '———',  # \u2014 em dash
        '<-->',
        '-->',
        '...',
        '--',
        '––',  # \u2013 en dash
        '——',  # \u2014 em dash
        '``',
        "''",
        '->',
        '<',
        '>',
        '–',  # \u2013 en dash
        '—',  # \u2014 em dash
        '―',  # \u2015 horizontal bar
        '~',  # \u007e Tilde
        '⁓',  # \u2053 Swung dash
        '∼',  # \u223c Tilde operator
        '°',  # \u00b0 Degrees
        ';',
        '@',
        '#',
        '$',
        '£',  # \u00a3
        '€',  # \u20ac
        '%',
        '&',
        '?',
        '!',
        '™',  # \u2122
        '®',  # \u00ae
        '…',  # \u2026
        '⋯',  # \u22ef Mid-line ellipsis
        '†',  # \u2020 Dagger
        '‡',  # \u2021 Double dagger
        '§',  # \u00a7 Section sign
        '¶'   # \u00b6 Pilcrow sign
        '≠',  # \u2260
        '≡',  # \u2261
        '≢',  # \u2262
        '≣',  # \u2263
        '≤',  # \u2264
        '≥',  # \u2265
        '≦',  # \u2266
        '≧',  # \u2267
        '≨',  # \u2268
        '≩',  # \u2269
        '≪',  # \u226a
        '≫',  # \u226b
        '≈',  # \u2248
        '=',
        '÷',  # \u00f7
        '×',  # \u00d7
        '→',  # \u2192
        '⇄',  # \u21c4
        '"',  # \u0022 Quote mark
        '“',  # \u201c
        '”',  # \u201d
        '„',  # \u201e
        '‟',  # \u201f
        '‘',  # \u2018 Left single quote
        '’',  # \u2019 Right single quote
        '‚',  # \u201a Single low quote
        '‛',  # \u201b Single reversed quote
        '`',  # \u0060
        '´',  # \u00b4
        # Primes
        '′',  # \u2032
        '″',  # \u2033
        '‴',  # \u2034
        '‵',  # \u2035
        '‶',  # \u2036
        '‷',  # \u2037
        '⁗',  # \u2057
        # Brackets
        '(',
        '[',
        '{',
        '}',
        ']',
        ')',
        # Slashes
        '/',  # \u002f Solidus
        '⁄',  # \u2044 Fraction slash
        '∕',  # \u2215 Division slash
        # Hyphens and Minuses
        '-',  # \u002d Hyphen-minus
        '−',  # \u2212 Minus
        '‒',  # \u2012 figure dash
        '‐',  # \u2010 Hyphen
        '‑',  # \u2011 Non-breaking hyphen
        '+',  # \u002b Plus
        '±',  # \u00b1 Plus/Minus
        ':',
    ]
    #: Split before these sequences if they end a token
    SPLIT_NO_DIGIT = [',']
    NO_SPLIT = {}
    #: Don't split around hyphens with these prefixes
    NO_SPLIT_PREFIX = {}
    #: Don't split around hyphens with these suffixes.
    NO_SPLIT_SUFFIX = {}

    def _subspan(self, s, span, nextspan):
        """Recursively subdivide spans based on a series of rules."""

        # Split on boundaries between greek and non-greek
        text = s[span[0]:span[1]]
        for i, char in enumerate(text):
            if i < len(text) - 1:
                nextchar = text[i + 1]
                if (char in GREEK and nextchar not in GREEK) or (char not in GREEK and nextchar in GREEK):
                    return [(span[0], span[0] + i + 1), (span[0] + i + 1, span[1])]

        # Perform all normal WordTokenizer splits
        return super(FineWordTokenizer, self)._subspan(s,span, nextspan)



# -*- coding: utf-8 -*-
"""
batterydataextractor.nlp.tag

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Tagger implementations. Used for part-of-speech tagging and named entity recognition.
"""
from abc import ABCMeta, abstractmethod
import six
from transformers import pipeline, AutoTokenizer


class BaseTagger(six.with_metaclass(ABCMeta)):
    """Abstract tagger class from which all taggers inherit.
    Subclasses must implement a ``tag()`` method.
    """

    @abstractmethod
    def tag(self, tokens):
        """Return a list of (token, tag) tuples for the given list of token strings.
        :param list(str) tokens: The list of tokens to tag.
        :rtype: list(tuple(str, str))
        """
        return

    def tag_sents(self, sentences):
        """Apply the ``tag`` method to each sentence in ``sentences``."""
        return [self.tag(s) for s in sentences]

    def evaluate(self, gold):
        """Evaluate the accuracy of this tagger using a gold standard corpus.
        :param list(list(tuple(str, str))) gold: The list of tagged sentences to score the tagger on.
        :returns: Tagger accuracy value.
        :rtype: float
        """
        tagged_sents = self.tag_sents([w for (w, t) in sent] for sent in gold)
        gold_tokens = sum(gold, [])
        test_tokens = sum(tagged_sents, [])
        accuracy = float(sum(x == y for x, y in six.moves.zip(gold_tokens, test_tokens))) / len(test_tokens)
        return accuracy


class NoneTagger(BaseTagger):
    """Tag every token with None."""

    def tag(self, tokens):
        """"""
        return [(token, None) for token in tokens]


class BertTagger(BaseTagger):
    """BERT Tagger"""

    def __init__(self, model=None, device=None):
        """"""
        self.model = model if model is not None else "batterydata/bde-pos-bert-cased-base"
        self.device = device if device is not None else -1
        self.tokenizer = AutoTokenizer.from_pretrained(self.model, model_max_length=512)

    def tag(self, tokens):
        """Return a list of (token, tag) tuples for a given list of (token, tag) tuples.

        :param list(str) tokens: The list of tokens to tag.
        """
        classifier = pipeline("token-classification", model=self.model, tokenizer=self.tokenizer,
                              aggregation_strategy="simple", device=self.device)
        tags = [token[0]['entity_group'] for token in classifier(tokens)]
        tagged_sent = list(zip(tokens, tags))
        return tagged_sent



# -*- coding: utf-8 -*-
"""
batterydataextractor.text.__init__

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Tools for processing text.
"""
import re
import unicodedata

from bs4 import UnicodeDammit

#: Control characters.
CONTROLS = {
    '\u0001', '\u0002', '\u0003', '\u0004', '\u0005', '\u0006', '\u0007', '\u0008', '\u000e', '\u000f', '\u0011',
    '\u0012', '\u0013', '\u0014', '\u0015', '\u0016', '\u0017', '\u0018', '\u0019', '\u001a', '\u001b',
}
# There are further control characters, but they are instead replaced with a space by unicode normalization
# '\u0009', '\u000a', '\u000b', '\u000c', '\u000d', '\u001c',  '\u001d', '\u001e', '\u001f'


#: Hyphen and dash characters.
HYPHENS = {
    '-',  # \u002d Hyphen-minus
    '‐',  # \u2010 Hyphen
    '‑',  # \u2011 Non-breaking hyphen
    '⁃',  # \u2043 Hyphen bullet
    '‒',  # \u2012 figure dash
    '–',  # \u2013 en dash
    '—',  # \u2014 em dash
    '―',  # \u2015 horizontal bar
}

#: Minus characters.
MINUSES = {
    '-',  # \u002d Hyphen-minus
    '−',  # \u2212 Minus
    '－',  # \uff0d Full-width Hyphen-minus
    '⁻',  # \u207b Superscript minus
}

#: Plus characters.
PLUSES = {
    '+',  # \u002b Plus
    '＋',  # \uff0b Full-width Plus
    '⁺',  # \u207a Superscript plus
}

#: Slash characters.
SLASHES = {
    '/',  # \u002f Solidus
    '⁄',  # \u2044 Fraction slash
    '∕',  # \u2215 Division slash
}

#: Tilde characters.
TILDES = {
    '~',  # \u007e Tilde
    '˜',  # \u02dc Small tilde
    '⁓',  # \u2053 Swung dash
    '∼',  # \u223c Tilde operator
    '∽',  # \u223d Reversed tilde
    '∿',  # \u223f Sine wave
    '〜',  # \u301c Wave dash
    '～',  # \uff5e Full-width tilde
}

#: Apostrophe characters.
APOSTROPHES = {
    "'",  # \u0027
    '’',  # \u2019
    '՚',  # \u055a
    'Ꞌ',  # \ua78b
    'ꞌ',  # \ua78c
    '＇',  # \uff07
}

#: Single quote characters.
SINGLE_QUOTES = {
    "'",  # \u0027
    '‘',  # \u2018
    '’',  # \u2019
    '‚',  # \u201a
    '‛',  # \u201b

}

#: Double quote characters.
DOUBLE_QUOTES = {
    '"',  # \u0022
    '“',  # \u201c
    '”',  # \u201d
    '„',  # \u201e
    '‟',  # \u201f
}

#: Accent characters.
ACCENTS = {
    '`',  # \u0060
    '´',  # \u00b4
}

#: Prime characters.
PRIMES = {
    '′',  # \u2032
    '″',  # \u2033
    '‴',  # \u2034
    '‵',  # \u2035
    '‶',  # \u2036
    '‷',  # \u2037
    '⁗',  # \u2057
}

#: Quote characters, including apostrophes, single quotes, double quotes, accents and primes.
QUOTES = APOSTROPHES | SINGLE_QUOTES | DOUBLE_QUOTES | ACCENTS | PRIMES

#: Uppercase and lowercase greek letters.
GREEK = {
    'Α',  # \u0391
    'Β',  # \u0392
    'Γ',  # \u0393
    'Δ',  # \u0394
    'Ε',  # \u0395
    'Ζ',  # \u0396
    'Η',  # \u0397
    'Θ',  # \u0398
    'Ι',  # \u0399
    'Κ',  # \u039a
    'Λ',  # \u039b
    'Μ',  # \u039c
    'Ν',  # \u039d
    'Ξ',  # \u039e
    'Ο',  # \u039f
    'Π',  # \u03a0
    'Ρ',  # \u03a1
    'Σ',  # \u03a3
    'Τ',  # \u03a4
    'Υ',  # \u03a5
    'Φ',  # \u03a6
    'Χ',  # \u03a7
    'Ψ',  # \u03a8
    'Ω',  # \u03a9
    'α',  # \u03b1
    'β',  # \u03b2
    'γ',  # \u03b3
    'δ',  # \u03b4
    'ε',  # \u03b5
    'ζ',  # \u03b6
    'η',  # \u03b7
    'θ',  # \u03b8
    'ι',  # \u03b9
    'κ',  # \u03ba
    'λ',  # \u03bb
    'μ',  # \u03bc
    'ν',  # \u03bd
    'ξ',  # \u03be
    'ο',  # \u03bf
    'π',  # \u03c0
    'ρ',  # \u03c1
    'σ',  # \u03c3
    'τ',  # \u03c4
    'υ',  # \u03c5
    'φ',  # \u03c6
    'χ',  # \u03c7
    'ψ',  # \u03c8
    'ω',  # \u03c9
}

#: Names of greek letters spelled out as words.
GREEK_WORDS = {
    'Alpha', 'Beta', 'Gamma', 'Delta', 'Epsilon', 'Zeta', 'Eta', 'Theta', 'Iota', 'Kappa', 'Lambda', 'Mu', 'Nu', 'Xi',
    'Omicron', 'Pi', 'Rho', 'Sigma', 'Tau', 'Upsilon', 'Phi', 'Chi', 'Psi', 'Omega', 'alpha', 'beta', 'gamma', 'delta',
    'epsilon', 'zeta', 'eta', 'theta', 'iota', 'kappa', 'lamda', 'mu', 'nu', 'xi', 'omicron', 'pi', 'rho', 'sigma',
    'tau', 'upsilon', 'phi', 'chi', 'psi', 'omega'
}

#: Words that should not be capitalized in titles.
SMALL = {
    'a', 'an', 'and', 'as', 'at', 'but', 'by', 'en', 'for', 'if', 'in', 'of', 'on', 'or', 'the', 'to', 'v', 'v', 'via',
    'vs', 'vs'
}

#: Words that should not be capitalized in names.
NAME_SMALL = {
    'abu', 'bon', 'bin', 'da', 'dal', 'de', 'del', 'der', 'de', 'di', u'dí', 'ibn', 'la', 'le', 'san', 'st', 'ste',
    'van', 'vel', 'von', 'y'
}

# This isn't every possible TLD, just the most common, to avoid false positives.
TLDS = {
    'aero', 'asia', 'biz', 'cat', 'com', 'coop', 'edu', 'eu', 'gov', 'info', 'int', 'jobs', 'mil', 'mobi', 'museum',
    'name', 'net', 'org', 'pro', 'tel', 'travel', 'xxx', 'ad', 'as', 'ar', 'au', 'br', 'bz', 'ca', 'cc', 'cd', 'co',
    'ch', 'cn', 'de', 'dj', 'es', 'fr', 'fm', 'it', 'io', 'jp', 'la', 'ly', 'me', 'ms', 'nl', 'no', 'nu', 'ru', 'sc',
    'se', 'sr', 'su', 'tk', 'tv', 'uk', 'us', 'ws'
}

#: A variety of numbers, spelled out as words.
NUMBERS = {
    'zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine', 'ten', 'eleven', 'twelve',
    'thirteen', 'fourteen', 'fifteen', 'sixteen', 'seventeen', 'eighteen', 'nineteen', 'twenty', 'thirty', 'forty',
    'fifty', 'sixty', 'seventy', 'eighty', 'ninety', 'hundred', 'thousand', 'million', 'billion', 'trillion'
}

#: Regular expression that matches email addresses.
EMAIL_RE = re.compile(r'([\w\-\.\+%]+@(\w[\w\-]+\.)+[\w\-]+)', re.I | re.U)
#: Regular expression that matches DOIs.
DOI_RE = re.compile(r'^10\.\d{4,9}/[-\._;()/:A-Z0-9]+$', re.U)
#: Regular expression that matches ISSNs.
ISSN_RE = re.compile(r'^\d{4}-\d{3}[\dX]$', re.U)
#: Regular expression that matches control characters not allowed in XML.
CONTROL_RE = re.compile('[^\u0020-\uD7FF\u0009\u000A\u000D\uE000-\uFFFD\u10000-\u10FFFF]+')


def get_encoding(input_string, guesses=None, is_html=False):
    """Return the encoding of a byte string. Uses bs4 UnicodeDammit.
    :param string input_string: Encoded byte string.
    :param list[string] guesses: (Optional) List of encoding guesses to prioritize.
    :param bool is_html: Whether the input is HTML.
    """
    converted = UnicodeDammit(input_string, override_encodings=[guesses] if guesses else [], is_html=is_html)
    return converted.original_encoding


def levenshtein(s1, s2, allow_substring=False):
    """
    Return the Levenshtein distance between two strings.
    The Levenshtein distance (a.k.a "edit difference") is the number of characters that need to be substituted,
    inserted or deleted to transform s1 into s2.
    Setting the `allow_substring` parameter to True allows s1 to be a
    substring of s2, so that, for example, "hello" and "hello there" would have a distance of zero.
    :param string s1: The first string
    :param string s2: The second string
    :param bool allow_substring: Whether to allow s1 to be a substring of s2
    :returns: Levenshtein distance.
    :type: int
    """
    len1, len2 = len(s1), len(s2)
    lev = []
    for i in range(len1 + 1):
        lev.append([0] * (len2 + 1))
    for i in range(len1 + 1):
        lev[i][0] = i
    for j in range(len2 + 1):
        lev[0][j] = 0 if allow_substring else j
    for i in range(len1):
        for j in range(len2):
            lev[i + 1][j + 1] = min(lev[i][j + 1] + 1, lev[i + 1][j] + 1, lev[i][j] + (s1[i] != s2[j]))
    return min(lev[len1]) if allow_substring else lev[len1][len2]


def bracket_level(text, open={'(', '[', '{'}, close={')', ']', '}'}):
    """Return 0 if string contains balanced brackets or no brackets."""
    level = 0
    for c in text:
        if c in open:
            level += 1
        elif c in close:
            level -= 1
    return level


def is_punct(text):
    for char in text:
        if not unicodedata.category(char).startswith('P'):
            return False
    else:
        return True


def is_ascii(text):
    for char in text:
        if ord(char) >= 128:
            return False
    else:
        return True


def like_url(text):
    if len(text) < 1:
        return False
    if text.startswith('http://'):
        return True
    elif text.startswith('www.') and len(text) >= 5:
        return True
    if len(text) < 2 or text[0] == '.' or text[-1] == '.' or '.' not in text:
        return False
    tld = text.rsplit('.', 1)[1].split(':', 1)[0]
    if tld.endswith('/'):
        return True
    if tld.isalpha() and tld in TLDS:
        return True
    return False


def like_number(text):
    text = text.replace(',', '').replace('.', '')
    if text.isdigit():
        return True
    if text.count('/') == 1:
        num, denom = text.split('/')
        if like_number(num) and like_number(denom):
            return True
    if text in NUMBERS:
        return True
    return False


def word_shape(text):
    prev_m = ''
    seq = 0
    shape = []
    for c in text:
        if c.isdigit():
            m = 'd'  # Digits
        elif c in GREEK:
            m = 'g'  # Greek letters
        elif c.isalpha():
            m = 'X' if c.isupper() else 'x'  # Uppercase or lowercase alphabetical
        elif c in QUOTES:
            m = "'"  # Quotes and apostrophes
        elif c in {':', ';'}:
            m = ':'  # Colons and semicolons
        elif c in {'!', '?', '.'}:
            m = '.'  # Sentence ends
        elif c in {'(', '[', '{', ')', ']', '}'}:
            m = 'b'  # Brackets
        elif c in {'°', '%'}:
            m = 'u'  # units
        elif c in {'■', '◼', '●', '▲', '○', '◆', '▼', '⧫', '△', '◇', '▽', '⬚', '□'}:
            m = 'l'  # list markers
        elif c in {',', '$', '&', '-'}:
            m = c  # Stay the same
        else:
            m = '*'  # Everything else, symbols etc: {'=', '+', '*', '_', '|', '@', '×', '÷', '±', '<', '≤', '>', '≥', '≦', '≡', '≅', '≈', '≃', '≲', '→', '←', '⇄', '≪', '≫', '↔', '≠', '∝', '∈', '⇌', '⇋', '⋯', '~', '·', '•', '√', '⊃', '∑', '∏', '®', '∞', '∂', '∫', '∇', '∧', '⟨', '⟩'}
        if m == prev_m:
            seq += 1
        else:
            seq = 0
            prev_m = m
        if seq < 3:
            shape.append(m)
    return ''.join(shape)



import unittest
from batterydataextractor.nlp import BertCemTagger, CemTagger
from batterydataextractor.doc import Document, Span


class TestBertCemTagger(unittest.TestCase):
    def test_false_pos(self):
        """Test the Chem BERT Tagger on a simple sentence."""
        dt = BertCemTagger()
        self.assertEqual(
            [
                (('UV-vis', 'JJ'), 'O'),
                (('spectrum', 'NN'), 'O'),
                (('of', 'IN'), 'O'),
                (('Coumarin', 'NN'), 'MAT')
            ],
            dt.tag([
                ('UV-vis', 'JJ'),
                ('spectrum', 'NN'),
                ('of', 'IN'),
                ('Coumarin', 'NN')
            ])
        )


class TestCemTagger(unittest.TestCase):
    """Test combined CemTagger."""

    def test_stoplist(self):
        """Test CemTagger removes words in stoplist, including words entirely made up of ignore prefix/suffix.
        GitHub issue #12.
        """
        ct = CemTagger()
        self.assertEqual([('LiFePO4', 'MAT')], ct.tag([('LiFePO4', 'NN')]))

    def test_cems_stoplist(self):
        """Test Document cems removes words in stoplist, ncluding words entirely made up of ignore prefix/suffix.
        GitHub issue #12.
        """
        self.assertEqual([Span('LiFePO4', 0, 7)], Document('LiFePO4 materials').cems)
        self.assertEqual([], Document('-aromatic').cems)
        self.assertEqual([], Document('non-aromatic').cems)


if __name__ == '__main__':
    unittest.main()



import logging
import unittest

from batterydataextractor.doc.document import Document
from batterydataextractor.doc.text import Heading, Paragraph

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestParseHeading(unittest.TestCase):

    maxDiff = None

    def do_parse(self, input, expected):
        s = Heading(input)
        log.debug(s)
        log.debug(s.tagged_tokens)
        results = [r.serialize() for r in s.records]
        log.debug(results)
        log.debug(expected)
        log.debug(results == expected)
        self.assertEqual(expected, results)

    def test_preparation_of(self):
        s = 'Preparation of 4-Amino-3-chloro-6-(2,3,4-trifluorophenyl)pyridine-2-carboxylic (Compound 41)'
        expected = [
            {'Compound': {'names': ['4-Amino-3-chloro-6-(2,3,4-trifluorophenyl)pyridine-2-carboxylic']}}
        ]
        self.do_parse(s, expected)

    def test_background(self):
        """Ensure uppercase heading isn't recognised (as a likely abbreviation)"""
        s = 'BACKGROUND'
        expected = []
        self.do_parse(s, expected)

    def test_background_art(self):
        """Ensure uppercase heading isn't recognised (as a likely abbreviation)"""
        s = 'BACKGROUND ART'
        expected = []
        self.do_parse(s, expected)

    def test_name1(self):
        s = '4-[4-(4-[5-[5-(4-Hydroxyphenyl)-3-phenyl-1H-pyrrol-2-ylimino]-4-phenyl-5H-pyrrol-2-yl]-phenoxymethyl)-[1,2,3]triazol-1-yl]butyric (8b)'
        expected = [{'Compound': {'names': ['4-[4-(4-[5-[5-(4-Hydroxyphenyl)-3-phenyl-1H-pyrrol-2-ylimino]-4-phenyl-5H-pyrrol-2-yl]-phenoxymethyl)-[1,2,3]triazol-1-yl]butyric']}}]
        self.do_parse(s, expected)

    def test_fluorescent_nano_beads(self):
        """"""
        s = 'Fluorescent Nano-Beads'
        expected = []
        self.do_parse(s, expected)

    def test_test(self):
        """"""
        s = 'Test 2'
        expected = []
        self.do_parse(s, expected)

    def test_name2(self):
        """"""
        s = '1-(3,4-Dibenzyloxycinnamoyl)-3,4′-dibenzyloxyresveratrol (14):'
        expected = [{'Compound': {'names': [u'1-(3,4-Dibenzyloxycinnamoyl)-3,4\u2032-dibenzyloxyresveratrol']}}]
        self.do_parse(s, expected)

    def test_name3(self):
        """"""
        s = '3.2 [3-(2-p-Tolylimidazo[1,2-a]pyridin-6-yl)phenyl]methanol'
        expected = [{'Compound': {'names': [u'[3-(2-p-Tolylimidazo[1,2-a]pyridin-6-yl)phenyl]methanol']}}]
        self.do_parse(s, expected)

    def test_name4(self):
        """"""
        s = 'Preparation of (E)-1-(4-(benzyloxy)phenyl)-2-(3,5-bis(benzyloxy)phenyl)ethene (I)'
        expected = [{'Compound': {'names': [u'(E)-1-(4-(benzyloxy)phenyl)-2-(3,5-bis(benzyloxy)phenyl)ethene']}}]
        self.do_parse(s, expected)

    def test_name5(self):
        """"""
        s = 'Preparation of 2-(10-bromoanthracene-9-yl)thiophene, 11'
        expected = [{'Compound': {'names': ['2-(10-bromoanthracene-9-yl)thiophene']}}]
        self.do_parse(s, expected)


class TestParseDocument(unittest.TestCase):

    maxDiff = None

    def test_consecutive_headings(self):
        d = Document(
            Heading('Preparation of 2-Amino-3-methoxy-5-chloropyridine'),
            Heading('Example 3'),
            Paragraph('The solid is suspended in hexanes, stirred and filtered to give the product as a bright yellow solid. (MP 93-94\xc2\xb0 C.).')
        )
        results = [r.serialize() for r in d.records]
        self.assertEqual(results, [{'Compound': {'names': [u'2-Amino-3-methoxy-5-chloropyridine']}},
                                   {'Compound': {'names': [u'hexanes']}}])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.parse.bert

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Bert parsers.
"""
import logging
import re
from abc import ABC

from .base import BaseSentenceParser
from transformers import AutoTokenizer
from transformers.pipelines import pipeline

log = logging.getLogger(__name__)


class BertParser(BaseSentenceParser, ABC):
    """Bert Parser"""

    def qa_model(self, model_name="batterydata/batterybert-cased-squad-v1") -> pipeline:
        return pipeline('question-answering', model=model_name, device=self.model.device,
                        tokenizer=AutoTokenizer.from_pretrained(model_name, model_max_length=512))


class BertMaterialParser(BertParser):
    """Bert Material Parser."""

    def interpret(self, tokens):
        bert_model = self.qa_model()
        context_list = [token[0] for token in tokens]
        context = " ".join(context_list)
        for specifier in self.model.defined_names:
            length = len(specifier.split(" "))
            if (specifier in context_list and length == 1) or (specifier in context and length > 1):
                question = "What is the value of {}?".format(specifier)
                qa_input = {'question': question, 'context': context}
                res = bert_model(qa_input, top_k=1)
                cs1 = res['score']
                if cs1 > self.model.confidence_threshold:
                    question2 = "What material has a {} of {}?".format(specifier, res['answer'])
                    qa_input2 = {'question': question2, 'context': context}
                    res2 = bert_model(qa_input2, top_k=1)
                    cs2 = res2['score']
                    value = re.findall(r'(?:\d*\.\d+|\d+)', res['answer'])
                    c = self.model(value=[float(v) for v in value],
                                   units=res['answer'].split(value[-1])[-1].strip(),
                                   raw_value=res['answer'],
                                   specifier=specifier,
                                   material=res2['answer'],
                                   confidence_score="%.4f" % (cs1 * cs2),
                                   original_text=context if self.model.original_text else None,
                                   )
                    yield c


class BertGeneralParser(BertParser):
    """Bert General Parser."""

    def interpret(self, tokens):
        bert_model = self.qa_model()
        context = " ".join([token[0] for token in tokens])
        for specifier in self.model.defined_names:
            question = specifier if self.model.self_defined else "What is the {}?".format(specifier)
            qa_input = {'question': question, 'context': context}
            res = bert_model(qa_input, top_k=1)
            if res['score'] > self.model.confidence_threshold:
                c = self.model(answer=res['answer'],
                               specifier=specifier,
                               confidence_score="%.4f" % res['score'],
                               original_text=context if self.model.original_text else None,
                               )
                yield c



# -*- coding: utf-8 -*-
"""
batterydataextractor.doc.meta

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
MetaData Document elements
"""
from .element import BaseElement
import logging
log = logging.getLogger(__name__)


class MetaData(BaseElement):

    def __init__(self, data):
        super(MetaData, self).__init__()
        self._data = data
        self._title = None
        self._authors = None
        self._publisher = None
        self._journal = None
        self._volume = None
        self._issue = None
        self._firstpage = None
        self._lastpage = None
        self._doi = None
        self._date = None
        self._language = None
        self._pdf_url = None
        self._html_url = None
        self._abstract = None
        for key, value in data.items():
            setattr(self, key, value)

    def __repr__(self):
        return {k: v for k, v in self.data.items() if v}.__str__()

    @property
    def records(self):
        return []

    def serialize(self):
        return {k: v for k, v in self.data.items() if v}

    @property
    def title(self):
        """The article title"""

        return self._title

    @property
    def authors(self):
        """The article Authors
        type:: list()
        """
        return self._authors

    @property
    def publisher(self):
        """The source publisher"""
        return self._publisher

    @property
    def journal(self):
        """The source journal"""
        return self._journal

    @property
    def volume(self):
        """The source volume"""
        return self._volume

    @property
    def issue(self):
        """The source issue"""
        return self._issue

    @property
    def firstpage(self):
        """The source first page title"""
        return self._firstpage

    @property
    def lastpage(self):
        """The source last page"""
        return self._lastpage

    @property
    def doi(self):
        """The source DOI"""
        return self._doi

    @property
    def pdf_url(self):
        """The source url to the PDF version"""
        return self._pdf_url

    @property
    def html_url(self):
        """The source url to the HTML version"""
        return self._html_url

    @property
    def abstract(self):
        """The abstract"""
        return self._abstract

    @property
    def date(self):
        """The source publish date"""
        return self._date

    @property
    def data(self):
        """Returns all data as a dict()"""
        return {k.lstrip('_'): v for k, v in self._data.items()}

    @property
    def abbreviation_definitions(self):
        return []



# -*- coding: utf-8 -*-
"""
batterydataextractor.reader.springer

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Springer XML reader
"""
from ..scrape.clean import clean, Cleaner
from ..scrape.springer import spr_clean_abstract, spr_clean_ref
from ..doc.meta import MetaData
from .markup import XmlReader

# XML stripper that removes the tags around numbers in chemical formulas
strip_spr_xml = Cleaner(strip_xpath='.//bold | .//label | .//xref | .//italic | .//sub | .//sup',
                        kill_xpath='.//tex-math | .//alternatives | .//mml:math | .//label')


class SpringerXmlReader(XmlReader):
    """Reader for Springer XML documents."""

    cleaners = [clean, spr_clean_abstract, spr_clean_ref, strip_spr_xml]

    root_css = 'article'
    title_css = 'article-title'
    heading_css = 'title'
    reference_css = 'xref'
    citation_css = 'ref'
    abstract_css = 'abstract'
    table_css = 'ce|table'
    figure_css = 'graphic'

    metadata_css = 'front'
    metadata_title_css = 'article-title'
    metadata_author_css = 'surname'
    metadata_journal_css = 'journal-title'
    metadata_volume_css = 'volume'
    metadata_issue_css = 'issue'
    metadata_publisher_css = 'publisher-name'
    metadata_day_css = 'day'
    metadata_month_css = 'month'
    metadata_year_css = 'year'
    metadata_firstpage_css = 'fpage'
    metadata_lastpage_css = 'lpage'
    metadata_abstract_css = 'abstract'
    metadata_doi_css = 'article-id[pub-id-type="doi"]'

    ignore_css = 'tex-math, label, table'

    def detect(self, fstring, fname=None):
        """Springer document detection based on string found in xml"""
        if fname and not fname.endswith('.xml'):
            return False
        if b'dtd-version="1.2"' in fstring:
            return True
        return False

    def _parse_metadata(self, el, refs, specials):
        title = self._css(self.metadata_title_css, el)
        authors = self._css(self.metadata_author_css, el)
        publisher = self._css(self.metadata_publisher_css, el)
        journal = self._css(self.metadata_journal_css, el)
        day = self._css(self.metadata_day_css, el)
        month = self._css(self.metadata_month_css, el)
        year = self._css(self.metadata_year_css, el)
        volume = self._css(self.metadata_volume_css, el)
        issue = self._css(self.metadata_issue_css, el)
        firstpage = self._css(self.metadata_firstpage_css, el)
        lastpage = self._css(self.metadata_lastpage_css, el)
        doi = self._css(self.metadata_doi_css, el)
        abstract = self._css(self.metadata_abstract_css, el)
        month = '0' + month[0].text if len(month[0].text) == 1 else month[0].text
        day = '0' + day[0].text if len(day[0].text) == 1 else day[0].text
        date = year[0].text + month + day
        try:
            abstract = abstract[0].findall('p')[0].text
        except IndexError:
            abstract = None

        metadata = {
                '_title': title[0].text if title else None,
                '_authors': [i.text for i in authors] if authors else None,
                '_publisher': publisher[0].text if publisher else None,
                '_journal': journal[0].text if journal else None,
                '_date': date if date else None,
                '_volume': volume[0].text if volume else None,
                '_issue': issue[0].text if issue else None,
                '_firstpage': firstpage[0].text if firstpage else None,
                '_lastpage': lastpage[0].text if lastpage else None,
                '_doi': doi[0].text if doi else None,
                '_abstract': abstract if abstract else None,
                }
        meta = MetaData(metadata)
        return [meta]



# -*- coding: utf-8 -*-
"""
batterydataextractor.scrape.base

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Base web-scraper
"""
from transformers import BertTokenizer, BertForSequenceClassification


class BaseWebScraper:
    """
    Base web-scraper.
    """
    def __init__(self, model_name_or_path="batterydata/batteryscibert-uncased-abstract"):
        """
        :param model_name_or_path: the BERT model to classify battery paper abstract. (Not required)
        """
        self.model_name_or_path = model_name_or_path
        self.tokenizer = BertTokenizer.from_pretrained(self.model_name_or_path)
        self.model = BertForSequenceClassification.from_pretrained(self.model_name_or_path)

    def classify_paper(self, abstract):
        """
        Battery or non-battery paper classifier
        :param abstract: abstract text from a scientific paper
        :return: predicted label. (0: non-battery paper; 1: battery paper)
        """
        inputs = self.tokenizer(abstract,  padding=True, truncation=True, max_length=512, return_tensors="pt")
        outputs = self.model(**inputs)
        probs = outputs[0].softmax(1)
        label = probs.argmax().item()
        return label



# -*- coding: utf-8 -*-
"""
batterydataextractor.text.normalize

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Tools for normalizing text.
"""
from abc import ABCMeta, abstractmethod
import re
import unicodedata

import six

from . import CONTROLS, HYPHENS, DOUBLE_QUOTES, ACCENTS, SINGLE_QUOTES, APOSTROPHES, SLASHES, TILDES, MINUSES
from .processors import BaseProcessor


class BaseNormalizer(six.with_metaclass(ABCMeta, BaseProcessor)):
    """Abstract normalizer class from which all normalizers inherit.
    Subclasses must implement a ``normalize()`` method.
    """

    @abstractmethod
    def normalize(self, text):
        """Normalize the text.
        :param string text: The text to normalize.
        :returns: Normalized text.
        :rtype: string
        """
        return text

    def __call__(self, text):
        """Calling a normalizer instance like a function just calls the normalize method."""
        return self.normalize(text)


class Normalizer(BaseNormalizer):
    """Main Normalizer class for generic English text.
    Normalize unicode, hyphens, quotes, whitespace.
    By default, the normal form NFKC is used for unicode normalization. This applies a compatibility decomposition,
    under which equivalent characters are unified, followed by a canonical composition. See Python docs for information
    on normal forms: http://docs.python.org/2/library/unicodedata.html#unicodedata.normalize
    """

    def __init__(self, form='NFKC', strip=True, collapse=True, hyphens=False, quotes=False, ellipsis=False,
                 slashes=False, tildes=False):
        """
        :param string form: Normal form for unicode normalization.
        :param bool strip: Whether to strip whitespace from start and end.
        :param bool collapse: Whether to collapse all whitespace (tabs, newlines) down to single spaces.
        :param bool hyphens: Whether to normalize all hyphens, minuses and dashes to the ASCII hyphen-minus character.
        :param bool quotes: Whether to normalize all apostrophes, quotes and primes to the ASCII quote character.
        :param bool ellipsis: Whether to normalize ellipses to three full stops.
        :param bool slashes: Whether to normalize slash characters to the ASCII slash character.
        :param bool tildes: Whether to normalize tilde characters to the ASCII tilde character.
        """
        self.form = form
        self.strip = strip
        self.collapse = collapse
        self.hyphens = hyphens
        self.quotes = quotes
        self.ellipsis = ellipsis
        self.slashes = slashes
        self.tildes = tildes

    def normalize(self, text):
        """Run the Normalizer on a string.
        :param text: The string to normalize.
        """
        # Normalize to canonical unicode (using NFKC by default)
        if self.form is not None:
            text = unicodedata.normalize(self.form, text)

        # Strip out any control characters (they occasionally creep in somehow)
        for control in CONTROLS:
            text = text.replace(control, '')

        # Normalize unusual whitespace not caught by unicodedata
        text = text.replace('\u000b', ' ').replace('\u000c', ' ').replace(u'\u0085', ' ')
        text = text.replace('\u2028', '\n').replace('\u2029', '\n').replace('\r\n', '\n').replace('\r', '\n')

        # Normalize all hyphens, minuses and dashes to ascii hyphen-minus and remove soft hyphen entirely
        if self.hyphens:
            for hyphen in HYPHENS | MINUSES:
                text = text.replace(hyphen, '-')
            text = text.replace('\u00ad', '')

        # Normalize all quotes and primes to ascii apostrophe and quotation mark
        if self.quotes:
            for double_quote in DOUBLE_QUOTES:
                text = text.replace(double_quote, '"')  # \u0022
            for single_quote in (SINGLE_QUOTES | APOSTROPHES | ACCENTS):
                text = text.replace(single_quote, "'")  # \u0027
            text = text.replace('′', "'")     # \u2032 prime
            text = text.replace('‵', "'")     # \u2035 reversed prime
            text = text.replace('″', "''")    # \u2033 double prime
            text = text.replace('‶', "''")    # \u2036 reversed double prime
            text = text.replace('‴', "'''")   # \u2034 triple prime
            text = text.replace('‷', "'''")   # \u2037 reversed triple prime
            text = text.replace('⁗', "''''")  # \u2057 quadruple prime

        if self.ellipsis:
            text = text.replace('…', '...').replace(' . . . ', ' ... ')  # \u2026

        if self.slashes:
            for slash in SLASHES:
                text = text.replace(slash, '/')

        if self.tildes:
            for tilde in TILDES:
                text = text.replace(tilde, '~')

        if self.strip:
            text = text.strip()

        # Collapse all whitespace down to a single space
        if self.collapse:
            text = ' '.join(text.split())
        return text

#: Default normalize that canonicalizes unicode and fixes whitespace.
normalize = Normalizer(strip=True, collapse=True, hyphens=False, quotes=False, ellipsis=False)
#: More aggressive normalize that also standardizes hyphens, and quotes.
strict_normalize = Normalizer(strip=True, collapse=True, hyphens=True, quotes=True, ellipsis=True, tildes=True)


# class ExcessNormalizer(Normalizer):
#     """Excessive string normalization.
#     This is useful when doing fuzzy string comparisons. A common use case is to run this before calculating the
#     Levenshtein distance between two strings, so that only "important" differences are counted.
#     """
#
#     def __init__(self, form='NFKC', strip=True, collapse=True, hyphens=True, quotes=True, ellipsis=True, tildes=True):
#         """"""
#         super(ExcessNormalizer, self).__init__(form, strip=strip, collapse=collapse, hyphens=hyphens, quotes=quotes,
#                                                ellipsis=ellipsis, tildes=tildes)
#
#     def normalize(self, text):
#         # Lowercase and normalize unicode
#         text = super(ExcessNormalizer, self).normalize(text.lower())
#         # Remove all whitespace
#         text = ''.join(text.split())
#         # Convert all apostrophes, quotes, accents, primes to single ascii apostrophe
#         for quote in QUOTES:
#             text = text.replace(quote, "'")
#         # Convert all brackets to regular parentheses
#         for ob in {'(', '<', '[', '{', '&lt;'}:
#             text = text.replace(ob, '(')
#         for cb in {')', '>', ']', '}', '&gt;'}:
#             text = text.replace(cb, '(')
#         return text


# excess_normalize = ExcessNormalizer(strip=True, collapse=True, hyphens=True, quotes=True, ellipsis=True, tildes=True)


class ChemNormalizer(Normalizer):
    """Normalizer that also unifies chemical spelling."""

    def __init__(self, form='NFKC', strip=True, collapse=True, hyphens=True, quotes=True, ellipsis=True, tildes=True,
                 chem_spell=True):
        """"""
        super(ChemNormalizer, self).__init__(form, strip=strip, collapse=collapse, hyphens=hyphens, quotes=quotes,
                                             ellipsis=ellipsis, tildes=tildes)
        self.chem_spell = chem_spell

    def normalize(self, text):
        """Normalize unicode, hyphens, whitespace, and some chemistry terms and formatting."""
        text = super(ChemNormalizer, self).normalize(text)
        # Normalize element spelling
        if self.chem_spell:
            text = re.sub(r'sulph', r'sulf', text, flags=re.I)
            text = re.sub(r'aluminum', r'aluminium', text, flags=re.I)
            text = re.sub(r'cesium', r'caesium', text, flags=re.I)
        return text

#
# chem_normalize = ChemNormalizer(strip=True, collapse=True, hyphens=True, quotes=True, ellipsis=True, tildes=True,
#                                 chem_spell=True)


# -*- coding: utf-8 -*-
"""
batterydataextractor.nlp.cem

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Named entity recognition (NER) for Chemical entity mentions (CEM).
"""
from .tag import BertTagger, BaseTagger


class BertCemTagger(BertTagger):
    """"""

    base_tagger = BertTagger()

    def tag(self, tokens):
        tuples = tokens
        cner_tagger = pipeline("token-classification", model="batterydata/bde-cner-batteryonlybert-uncased-base",
                               tokenizer=AutoTokenizer.from_pretrained("batterydata/bde-cner-batteryonlybert-uncased-base",
                                                                       model_max_length=512),
                               aggregation_strategy="simple", device=self.device)
        result = cner_tagger([token[0] for token in tuples])
        labels = ['O' if token == [] else 'MAT' for token in result]
        tagged_sent = list(zip(tuples, labels))
        return tagged_sent


class CemTagger(BaseTagger):
    """Return the combined output of a number of chemical entity taggers."""

    taggers = [BertCemTagger()]

    def tag(self, tokens):
        """Run individual chemical entity mention taggers and return union of matches, with some postprocessing."""
        # Combine output from individual taggers
        for tagger in self.taggers:
            tag_gen = tagger.tag(tokens)
            materials = [tag[0][0] for tag in tag_gen]
            tags = [tag[-1] for tag in tag_gen]
        token_tags = list(zip(materials, tags))
        return token_tags



# -*- coding: utf-8 -*-
"""
batterydataextractor.parse.base

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Base classes for parsing sentences and tables.
"""
from abc import abstractmethod, ABC
import logging

log = logging.getLogger(__name__)


class BaseParser:
    """"""
    model = None

    @abstractmethod
    def interpret(self, tokens):
        pass


class BaseSentenceParser(BaseParser, ABC):
    """
    Base class for parsing sentences. To implement a parser for a new property,
    implement the interpret function.
    """

    def parse_sentence(self, tokens):
        """
        Parse a sentence. This function is primarily called by the
        :attr:`~batterydataextractor.doc.text.Sentence.records` property of
        :class:`~batterydataextractor.doc.text.Sentence`.
        :param list[(token,tag)] tokens: List of tokens for parsing. When this method
            is called by :attr:`batterydataextractor.doc.text.Sentence.records`,
            the tokens passed in are :attr:`batterydataextractor.doc.text.Sentence.tagged_tokens`.
        :returns: All the models found in the sentence.
        :rtype: Iterator[:class:`batterydataextractor.model.base.BaseModel`]
        """
        for model in self.interpret(tokens):
            yield model



# -*- coding: utf-8 -*-
"""
batterydataextractor.reader.__init__

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Paper reader
"""
from .elsevier import ElsevierXmlReader
from .markup import HtmlReader, XmlReader
from .plaintext import PlainTextReader
from .rsc import RscHtmlReader
from .springer import SpringerXmlReader


DEFAULT_READERS = [
    RscHtmlReader(),
    ElsevierXmlReader(),
    SpringerXmlReader(),
    XmlReader(),
    HtmlReader(),
    PlainTextReader(),
]



# -*- coding: utf-8 -*-
"""
batterydataextractor.config

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Config file reader/writer.
"""
import io
import os
import sys
from collections import MutableMapping

import appdirs
import yaml
from yaml import SafeLoader


def construct_yaml_str(self, node):
    """Override the default string handling function to always return unicode objects."""
    return self.construct_scalar(node)


SafeLoader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)


class Config(MutableMapping):
    """Read and write to config file.
    A config object is essentially a string key-value store that can be treated like a dictionary::
        c = Config()
        c['foo'] = 'bar'
        print c['foo']
    The file location may be specified::
        c = Config('~/matt/anotherconfig.yml')
        c['where'] = 'in a different file'
    If no location is specified, the environment variable BATTERYDATAEXTRACTOR_CONFIG is checked and used if available.
    Otherwise, a standard config location is used, which varies depending on the operating system. You can check the
    location using the ``path`` property. For more information see https://github.com/ActiveState/appdirs
    It is possible to edit the file by hand with a text editor. It is in YAML format.
    Warning: multiple instances of Config() pointing to the same file will not see each others' changes, and will
    overwrite the entire file when any key is changed.
    """

    def __init__(self, path=None):
        """
        :param string path: (Optional) Path to config file location.
        """
        self._path = path
        self._data = {}

        # Use BATTERYDATAEXTRACTOR_CONFIG environment variable if set
        if not self._path:
            self._path = os.environ.get('BATTERYDATAEXTRACTOR_CONFIG')
        # Use OS-dependent config directory given by appdirs
        if not self._path:
            if sys.version_info[0] == 2:
                self._path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'python2_config.yml')
            else:
                self._path = os.path.join(appdirs.user_config_dir('BatteryDataExtractor'), 'batterydataextractor.yml')
        if os.path.isfile(self.path):
            with io.open(self.path, encoding='utf8') as f:
                self._data = yaml.safe_load(f)
                if self._data is None:
                    self._data = {}

    @property
    def path(self):
        """The path to the config file."""
        return self._path

    def _flush(self):
        """Save the contents of data to the file on disk. You should not need to call this manually."""
        d = os.path.dirname(self.path)
        if not os.path.isdir(d):
            os.makedirs(d)
        with io.open(self.path, 'w', encoding='utf8') as f:
            yaml.safe_dump(self._data, f, default_flow_style=False, encoding=None)

    def __contains__(self, k):
        return k in self._data

    def __getitem__(self, k):
        return self._data[k]

    def __setitem__(self, k, v):
        self._data[k] = v
        self._flush()

    def __delitem__(self, k):
        del self._data[k]
        self._flush()

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return '<Config: %s>' % self.path

    def clear(self):
        """Clear all values from config."""
        self._data = {}
        self._flush()


#: Global config instance.
config = Config()



import logging
import unittest
from batterydataextractor.nlp import ChemSentenceTokenizer
from batterydataextractor.doc import Text


logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestChemSentenceTokenizer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ps = ChemSentenceTokenizer()

    def test_sentence_tokenizer(self):
        """Test the Sentence Tokenizer directly."""
        text = 'This is the first sentence. This is another. This is a third.'
        sents = ['This is the first sentence.', 'This is another.', 'This is a third.']
        self.assertEqual(sents, self.ps.tokenize(text))

    def test_et_al(self):
        """Test the tokenizer handles et al. within a sentence correctly."""
        text = 'Costa et al. reported the growth of HA nanowires due to the chemical potential of an amorphous calcium phosphate solution. This structural feature would make the {001} very sensitive to surrounding growth conditions.'
        sents = [
            'Costa et al. reported the growth of HA nanowires due to the chemical potential of an amorphous calcium phosphate solution.',
            'This structural feature would make the {001} very sensitive to surrounding growth conditions.'
        ]
        self.assertEqual(sents, self.ps.tokenize(text))

    def test_et_al2(self):
        """Test the tokenizer handles et al. abbreviation correctly."""
        text = 'In the field of DSCs, Gratzel et al. demonstrated this for the first time in 2004.'
        sents = ['In the field of DSCs, Gratzel et al. demonstrated this for the first time in 2004.']
        self.assertEqual(sents, self.ps.tokenize(text))

    def test_fig_bracket(self):
        """Test the tokenizer handles bracketed Fig. abbreviation correctly."""
        text = 'The model is in good agreement with the intensity of the peaks observed in the XRD patterns (Fig. 1).'
        sents = ['The model is in good agreement with the intensity of the peaks observed in the XRD patterns (Fig. 1).']
        self.assertEqual(sents, self.ps.tokenize(text))

    def test_eg_et_al(self):
        """Test the tokenizer handles e.g. and et al. abbreviations correctly."""
        text = 'There is a clear linkage to some diseases, e.g. multiple myeloma. Vidler et al. studied the druggability of the different members, but to date there are no PCM studies performed on this family.'
        sents = [
            'There is a clear linkage to some diseases, e.g. multiple myeloma.',
            'Vidler et al. studied the druggability of the different members, but to date there are no PCM studies performed on this family.'
        ]
        self.assertEqual(sents, self.ps.tokenize(text))

    def test_colon(self):
        """Test the tokenizer handles colons correctly."""
        text = 'The authors were able to split this into two types, namely: those involved in absorption and those involved in emission.'
        sents = ['The authors were able to split this into two types, namely: those involved in absorption and those involved in emission.']
        self.assertEqual(sents, self.ps.tokenize(text))

    def test_lowercase_sentence_start(self):
        """Test the tokenizer handles lowercase sentence start correctly."""
        text = 'These regions are positive contributors to overall efficiency. van Westen et al. built on this by including data from 24 new sources.'
        sents = [
            'These regions are positive contributors to overall efficiency.',
            'van Westen et al. built on this by including data from 24 new sources.'
        ]
        self.assertEqual(sents, self.ps.tokenize(text))

    def test_chemtext_sentence(self):
        """Test sentence tokenization through the ChemText and Sentence API."""
        t = Text('These regions are positive contributors to overall efficiency. van Westen et al. built on this by including data from 24 new sources.')
        self.assertEqual(
            [(0, 62, 'These regions are positive contributors to overall efficiency.'), (63, 133, 'van Westen et al. built on this by including data from 24 new sources.')],
            [(s.start, s.end, s.text) for s in t.sentences]
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.nlp.abbrev

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Abbreviation detection.
"""
from transformers import pipeline, AutoTokenizer
from abc import ABCMeta
import six


class AbbreviationDetector(six.with_metaclass(ABCMeta)):
    """"""

    def __init__(self, model_name="batterydata/bde-abbrev-batteryonlybert-cased-base", device=None):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, model_max_length=512)
        self.device = device if device else -1
        self.model = pipeline('token-classification', model_name, tokenizer=self.tokenizer,
                              aggregation_strategy='simple', device=self.device)

    def detect_spans(self, tokens):
        """
        Detects abbreviations in a list of tokens.

        :param tokens: a list of tokens
        :return:
        """
        results = self.model(" ".join(tokens))
        long, short = [], []
        for entity in results:
            if entity['entity_group'] == 'long':
                long.append((entity['start'], entity['end']))
            elif entity['entity_group'] == 'short':
                short.append((entity['start'], entity['end']))
        return short, long

    def detect(self, tokens):
        short_words = []
        long_words = []
        doc = " ".join(tokens)
        abbr_span, long_span = self.detect_spans(tokens)
        for abbr in abbr_span:
            short_words.append(("Abbr: ", doc[abbr[0]: abbr[1]]))
        for long in long_span:
            long_words.append(("LF: ", doc[long[0]: long[1]]))
        return short_words, long_words



# -*- coding: utf-8 -*-
"""
batterydataextractor.model.model

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Model classes for physical properties.
"""
import logging

from .base import BaseModel, StringType, ListType, FloatType

from ..parse.cem import CompoundParser
from ..parse.bert import BertMaterialParser, BertGeneralParser

log = logging.getLogger(__name__)


class Compound(BaseModel):
    names = ListType(StringType())
    parsers = [CompoundParser()]

    def merge(self, other):
        """Merge data from another Compound into this Compound."""
        log.debug('Merging: %s and %s' % (self.serialize(), other.serialize()))
        for k in self.keys():
            for new_item in other[k]:
                if new_item not in self[k]:
                    self[k].append(new_item)
        log.debug('Result: %s' % self.serialize())
        return self

    @property
    def is_unidentified(self):
        if not self.names:
            return True
        return False

    @property
    def is_id_only(self):
        """Return True if identifier information only."""
        for key, value in self.items():
            if key not in {'names', 'labels', 'roles'} and value:
                return False
        if self.names:
            return True
        return False


class PropertyData(BaseModel):
    value = ListType(FloatType(required=True))
    units = StringType(contextual=False)
    raw_value = StringType(contextual=False)
    specifier = StringType(contextual=False)
    material = StringType(contextual=False, required=True)
    confidence_score = FloatType(contextual=False)
    original_text = StringType(contextual=False)
    device = FloatType(contextual=False)
    parsers = [BertMaterialParser()]


class GeneralInfo(BaseModel):
    answer = StringType(contextual=False, required=True)
    specifier = StringType(contextual=False)
    confidence_score = FloatType(contextual=False)
    original_text = StringType(contextual=False)
    device = FloatType(contextual=False)
    parsers = [BertGeneralParser()]



# -*- coding: utf-8 -*-
"""
batterydataextractor.reader.clean

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Tools for cleaning up XML/HTML by removing tags entirely or replacing with their contents.
"""
import copy
import logging
import re
from lxml.etree import fromstring, tostring
from lxml.html import fromstring as html_fromstring
import six

log = logging.getLogger(__name__)

BLOCK_ELEMENTS = {
    'p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'ul', 'ol', 'pre', 'dd', 'dl', 'div', 'noscript', 'blockquote', 'form',
    'hr', 'table', 'fieldset', 'address', 'article', 'aside', 'audio', 'canvas', 'figcaption', 'figure', 'footer',
    'header', 'hgroup', 'output', 'section', 'body', 'head', 'title', 'tr', 'td', 'th', 'thead', 'tfoot', 'dt', 'li',
    'tbody',
}


class Cleaner(object):
    """Clean HTML or XML by removing tags completely or replacing with their contents.
    A Cleaner instance provides a ``clean_markup`` method::
        cleaner = Cleaner()
        htmlstring = '<html><body><script>alert("test")</script><p>Some text</p></body></html>'
        print(cleaner.clean_markup(htmlstring))
    A Cleaner instance is also a callable that can be applied to lxml document trees::
        tree = lxml.etree.fromstring(htmlstring)
        cleaner(tree)
        print(lxml.etree.tostring(tree))
    Elements that are matched by ``kill_xpath`` are removed entirely, along with their contents. By default,
    ``kill_xpath`` matches all script and style tags, as well as comments and processing instructions.
    Elements that are matched by ``strip_xpath`` are replaced with their contents. By default, no elements are stripped.
    A common use-case is to set ``strip_xpath`` to ``.//*``, which specifies that all elements should be stripped.
    Elements that are matched by ``allow_xpath`` are excepted from stripping, even if they are also matched by
    ``strip_xpath``. This is useful when setting ``strip_xpath`` to strip all tags, allowing a few expections to be
    specified by ``allow_xpath``.
    """

    kill_xpath = './/script | .//style | .//comment() | .//processing-instruction() | .//*[@style="display:none;"]'
    strip_xpath = None
    allow_xpath = None
    fix_whitespace = True

    namespaces = {
        're': 'http://exslt.org/regular-expressions',
        'set': 'http://exslt.org/sets',
        'dc': 'http://purl.org/dc/elements/1.1/',
        'prism': 'http://prismstandard.org/namespaces/basic/2.0/',
        'xml': 'http://www.w3.org/XML/1998/namespace',
        'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns',
    }

    def __init__(self, **kwargs):
        """Behaviour can be customized by overriding attributes in a subclass or setting them in the constructor.
        :param string kill_xpath: XPath expression for tags to remove along with their contents.
        :param string strip_xpath: XPath expression for tags to replace with their contents.
        :param string allow_xpath: XPath expression for tags to except from strip_xpath.
        :param bool fix_whitespace: Normalize whitespace to a single space and ensure newlines around block elements.
        :param dict namespaces: Namespace prefixes to register for the XPaths.
        """
        for name, value in kwargs.items():
            if not hasattr(self, name):
                raise TypeError('Unknown parameter: %s=%r' % (name, value))
            setattr(self, name, value)

    def __call__(self, doc):
        """Clean the document."""
        if hasattr(doc, 'getroot'):
            doc = doc.getroot()

        if self.fix_whitespace:
            # Ensure newlines around block elements
            for el in doc.iterdescendants():
                if el.tag in BLOCK_ELEMENTS:
                    el.tail = (el.tail or '') + '\n'
                    previous = el.getprevious()
                    parent = el.getparent()
                    if previous is None:
                        parent.text = (parent.text or '') + '\n'
                    else:
                        previous.tail = (previous.tail or '') + '\n'

        # Remove elements that match kill_xpath
        if self.kill_xpath:
            for el in doc.xpath(self.kill_xpath, namespaces=self.namespaces):
                #log.debug('Killing: %s' % tostring(el))
                parent = el.getparent()
                # We can't kill the root element!
                if parent is None:
                    continue
                if el.tail:
                    previous = el.getprevious()
                    if previous is None:
                        parent.text = (parent.text or '') + el.tail
                    else:
                        previous.tail = (previous.tail or '') + el.tail
                parent.remove(el)

        # Collect all the allowed elements
        to_keep = [el for el in doc.xpath(self.allow_xpath, namespaces=self.namespaces)] if self.allow_xpath else []

        # Replace elements that match strip_xpath with their contents
        if self.strip_xpath:
            for el in doc.xpath(self.strip_xpath, namespaces=self.namespaces):
                # Skip if allowed by allow_xpath
                if el in to_keep:
                    continue
                parent = el.getparent()
                previous = el.getprevious()
                # We can't strip the root element!
                if parent is None:
                    continue
                # Append the text to previous tail (or parent text if no previous), ensuring newline if block level
                if el.text and isinstance(el.tag, six.string_types):
                    if previous is None:
                        parent.text = (parent.text or '') + el.text
                    else:
                        previous.tail = (previous.tail or '') + el.text
                # Append the tail to last child tail, or previous tail, or parent text, ensuring newline if block level
                if el.tail:
                    if len(el):
                        last = el[-1]
                        last.tail = (last.tail or '') + el.tail
                    elif previous is None:
                        parent.text = (parent.text or '') + el.tail
                    else:
                        previous.tail = (previous.tail or '') + el.tail
                index = parent.index(el)
                parent[index:index+1] = el[:]

        # Collapse whitespace down to a single space or a single newline
        if self.fix_whitespace:
            for el in doc.iter():
                if el.text is not None:
                    el.text = re.sub(r'\s*\n\s*', '\n', el.text)
                    el.text = re.sub(r'[ \t]+', ' ', el.text)
                    # el.text = re.sub(r'\s+', ' ', el.text)
                if el.tail is not None:
                    el.tail = re.sub(r'\s*\n\s*', '\n', el.tail)
                    el.tail = re.sub(r'[ \t]+', ' ', el.tail)
                    # el.tail = re.sub(r'\s+', ' ', el.tail)

    def clean_html(self, html):
        """Apply ``Cleaner`` to HTML string or document and return a cleaned string or document."""
        result_type = type(html)
        if isinstance(html, six.string_types):
            doc = html_fromstring(html)
        else:
            doc = copy.deepcopy(html)
        self(doc)
        if issubclass(result_type, six.binary_type):
            return tostring(doc, encoding='utf-8')
        elif issubclass(result_type, six.text_type):
            return tostring(doc, encoding='unicode')
        else:
            return doc

    def clean_markup(self, markup, parser=None):
        """Apply ``Cleaner`` to markup string or document and return a cleaned string or document."""
        result_type = type(markup)
        if isinstance(markup, six.string_types):
            doc = fromstring(markup, parser=parser)
        else:
            doc = copy.deepcopy(markup)
        self(doc)
        if issubclass(result_type, six.binary_type):
            return tostring(doc, encoding='utf-8')
        elif issubclass(result_type, six.text_type):
            return tostring(doc, encoding='unicode')
        else:
            return doc


#: A default Cleaner instance, which kills comments, processing instructions, script tags, style tags.
clean = Cleaner()

#: Convenience function for applying ``clean`` to a string.
clean_markup = clean.clean_markup

#: Convenience function for applying ``clean`` to a HTML string.
clean_html = clean.clean_html

#: A Cleaner instance that is configured to strip all tags, replacing them with their text contents.
strip = Cleaner(strip_xpath='.//*')

#: Convenience function for applying ``strip`` to a string.
strip_markup = strip.clean_markup

#: Convenience function for applying ``strip`` to a HTML string.
strip_html = strip.clean_html



# -*- coding: utf-8 -*-
"""
batterydataextractor.doc.element

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Document elements.
"""
from abc import ABCMeta, abstractmethod
import json
import operator
import six
import csv

from ..model.model import PropertyData, GeneralInfo

import logging
log = logging.getLogger(__name__)


class BaseElement(six.with_metaclass(ABCMeta)):
    """
    Abstract base class for a Document Element.
    :ivar id: (Optional) An identifier for this Element.
    :type id: Any or None
    :ivar list[batterydataextractor.models.BaseModel] models: A list of models that this element will parse
    """

    def __init__(self, document=None, references=None, id=None, models=None, device=None, **kwargs):
        """
        .. note::
            If intended as part of a :class:`~batterydataextractor.doc.document.Document`,
            an element should either be initialized with a reference to its containing document,
            or its :attr:`document` attribute should be set as soon as possible.
            If the element is being passed in to a :class:`~batterydataextractor.doc.document.Document`
            to initialise it, the :attr:`document` attribute is automatically set
            during the initialisation of the document, so the user does not need to worry about this.
        :param Document document: (Optional) The document containing this element.
        :param list[Citation] references: (Optional) Any references contained in the element.
        :param Any id: (Optional) An identifier for this element. Must be equatable.
        :param list[batterydataextractor.models.BaseModel] models: (Optional) A list of models for this element to parse.
            If the element is part of another element (e.g. a :class:`~batterydataextractor.doc.text.Sentence`
            inside a :class:`~batterydataextractor.doc.text.Paragraph`), or is part of a :class:`batterydataextractor.doc.document.Document`,
            this is set automatically to be the same as that of the containing element, unless manually set otherwise.
        """
        #: The containing Document
        self._document = document
        self.id = id
        self.references = references if references is not None else []
        if models:
            self.models = models
        else:
            self.models = []
        self._streamlined_models_list = None

        if device:
            self._device = device
        else:
            self._device = -1

    def __repr__(self):
        return '<%s>' % (self.__class__.__name__,)

    def __str__(self):
        return '<%s>' % (self.__class__.__name__,)

    @property
    def document(self):
        """ The :class:`batterydataextractor.doc.document.Document` that this element belongs to. """
        return self._document

    @document.setter
    def document(self, document):
        # Subclasses may need to override this and also assign the document to sub-elements
        self._document = document

    @property
    @abstractmethod
    def records(self):
        """All records found in this Document, as a list of :class:`batterydataextractor.model.base.BaseModel`."""
        return []

    # @abstractmethod  # TODO: Put this back?
    # def serialize(self):
    #     """Convert Element to python dictionary."""
    #     return []

    def add_models(self, models):
        """Set all models on this element
        """
        log.debug("Setting models on %s" % self)
        self._streamlined_models_list = None
        self.models.extend(models)
        self.models = self.models

    def add_models_by_names(self, names, confidence_threshold=0, original_text=False):
        """"""
        model = PropertyData
        model.defined_names = names
        model.confidence_threshold = confidence_threshold
        model.original_text = original_text
        model.device = self.device
        self.models.extend([model])

    def add_general_models(self, names, confidence_threshold=0, original_text=False, self_defined=False):
        """"""
        model = GeneralInfo
        model.defined_names = names
        model.confidence_threshold = confidence_threshold
        model.original_text = original_text
        model.self_defined = self_defined
        model.device = self.device
        self.models.extend([model])

    @property
    def models(self):
        return self._models

    @models.setter
    def models(self, value):
        self._models = value
        self._streamlined_models_list = None

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, value):
        self._device = value

    @property
    def _streamlined_models(self):
        if self._streamlined_models_list is None:
            models = set()
            log.debug(self.models)
            for model in self.models:
                models.update(self._flatten_model(model))
            self._streamlined_models_list = sorted(list(models),
                                                   key=operator.attrgetter('__name__'))
        for model in self._streamlined_models_list:
            for parser in model.parsers:
                parser.model = model
        return self._streamlined_models_list

    def _flatten_model(self, model):
        model_set = {model}
        for field_name, field in six.iteritems(model.fields):
            if hasattr(field, 'model_class'):
                model_set.update(self._flatten_model(field.model_class))
        log.debug(model_set)
        return model_set

    def to_json(self, *args, **kwargs):
        """Convert element to JSON string. The content of the JSON will be equivalent
        to that of :meth:`serialize`."""
        return json.dumps(self.serialize(), *args, **kwargs)

    def to_database(self, file_name, file_type='json'):
        """
        Save the document to a file in the database.

        :param file_name: The name of the file to save the document to.
        :param file_type: The type of file to save the document to.
        """
        if file_type == 'json':
            with open(file_name, 'a', encoding='utf-8') as f:
                json.dump(self.serialize(), f)
                f.write('\n')
        elif file_type == 'txt':
            with open(file_name, 'a', encoding='utf-8') as f:
                f.write(json.dumps(self.serialize()))
                f.write('\n')
        elif file_type == 'csv':
            data = self.serialize()
            dic = data[list(data.keys())[0]]
            dic['model_type'] = list(dic.keys())[0]
            csv_columns = dic.keys()
            with open(file_name, 'a', encoding='utf-8') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=csv_columns)
                writer.writeheader()
                writer.writerow(dic)
        else:
            raise ValueError('Unknown file type. Please use json, txt, or csv.')


class CaptionedElement(BaseElement):
    """
    Document Element with a caption.
    :ivar BaseElement caption: The caption for this element.
    """

    def __init__(self, caption, label=None, **kwargs):
        """
        .. note::
            If intended as part of a :class:`~batterydataextractor.doc.document.Document`,
            an element should either be initialized with a reference to its containing document,
            or its :attr:`document` attribute should be set as soon as possible.
            If the element is being passed in to a :class:`~batterydataextractor.doc.document.Document`
            to initialise it, the :attr:`document` attribute is automatically set
            during the initialisation of the document, so the user does not need to worry about this.
        :param BaseElement caption: The caption for the element.
        :param Document document: (Optional) The document containing this element.
        :param str label: (Optional) The label for the captioned element, e.g. Table 1 would have a label of 1.
        :param Any id: (Optional) Some identifier for this element. Must be equatable.
        :param list[batterydataextractor.models.BaseModel] models: (Optional) A list of models for this element to parse.
            If the element is part of another element (e.g. a :class:`~batterydataextractor.doc.text.Sentence`
            inside a :class:`~batterydataextractor.doc.text.Paragraph`), or is part of a :class:`~batterydataextractor.doc.document.Document`,
            this is set automatically to be the same as that of the containing element, unless manually set otherwise.
        """
        # TODO: docs for label
        self.caption = caption
        self.label = label
        super(CaptionedElement, self).__init__(**kwargs)

    def __repr__(self):
        return '%s(id=%r, references=%r, caption=%r)' % (self.__class__.__name__, self.id, self.references, self.caption.text)

    def __str__(self):
        return self.caption.text

    @property
    def document(self):
        """ The :class:`~batterydataextractor.doc.document.Document` that this element belongs to. """
        return self._document

    @document.setter
    def document(self, document):
        self._document = document
        self.caption.document = document

    @property
    def records(self):
        """All records found in the object, as a list of :class:`~batterydataextractor.model.base.BaseModel`."""
        # This just passes the caption records. Subclasses may wish to extend this.
        return self.caption.records

    @property
    def abbreviation_definitions(self):
        """
        A list of all abbreviation definitions in this Document. Each abbreviation is in the form
        (:class:`str` abbreviation, :class:`str` long form of abbreviation, :class:`str` ner_tag)
        """
        return self.caption.abbreviation_definitions

    @property
    def ner_tags(self):
        """
        A list of all Named Entity Recognition tags in the caption for this element.
        If a word was found not to be a named entity, the named entity tag is None,
        and if it was found to be a named entity, it can have either a tag of 'B-CM' for a beginning of a
        mention of a chemical or 'I-CM' for the continuation of a mention.
        """
        # TODO: Delete this method?
        return self.caption.ner_tags

    @property
    def cems(self):
        """
        A list of all Chemical Entity Mentions in this document as :class:`~batterydataextractor.doc.text.Span`
        """
        return self.caption.cems

    @property
    def models(self):
        return self._models

    @models.setter
    def models(self, value):
        self._models = value
        self.caption.models = value

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, value):
        self._device = value
        self.caption.device = value

    def serialize(self):
        """
        Convert self to a dictionary. The key 'type' will contain
        the name of the class being serialized, and the key 'caption' will contain
        a serialized representation of :attr:`caption`, which is a :class:`~batterydataextractor.doc.element.BaseElement`
        """
        data = {'type': self.__class__.__name__, 'caption': self.caption.serialize()}
        return data



# -*- coding: utf-8 -*-
"""
test_reader_markup
~~~~~~~~~~~~~~~~~~
Test XML/HTML reader.
"""
import logging
import unittest

from batterydataextractor.doc import Paragraph
from batterydataextractor.reader import HtmlReader


logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestReaderHtml(unittest.TestCase):

    maxDiff = None

    def test_paragraphs(self):
        """Test paragraphs are correctly split."""
        r = HtmlReader()
        d = r.parse('<p>First para</p><p>Second Para</p>')
        self.assertEqual(len(d.elements), 2)
        self.assertEqual(d.elements[0].text, 'First para')
        self.assertEqual(d.elements[1].text, 'Second Para')
        for el in d.elements:
            self.assertIsInstance(el, Paragraph)

    def test_paragraphs2(self):
        """Test paragraphs are correctly split with no closing element."""
        r = HtmlReader()
        d = r.parse('<p>First para<p>Second Para')
        self.assertEqual(len(d.elements), 2)
        self.assertEqual(d.elements[0].text, 'First para')
        self.assertEqual(d.elements[1].text, 'Second Para')
        for el in d.elements:
            self.assertIsInstance(el, Paragraph)

    def test_linebreak(self):
        """Test br splits paragraph."""
        r = HtmlReader()
        d = r.parse('First line<br/>Second line')
        self.assertEqual(len(d.elements), 2)
        self.assertEqual(d.elements[0].text, 'First line')
        self.assertEqual(d.elements[1].text, 'Second line')
        for el in d.elements:
            self.assertIsInstance(el, Paragraph)

    def test_linebreak2(self):
        """Test br splits paragraph."""
        r = HtmlReader()
        d = r.parse('<span>First line</span><br/><span>Second line</span>')
        self.assertEqual(len(d.elements), 2)
        self.assertEqual(d.elements[0].text, 'First line')
        self.assertEqual(d.elements[1].text, 'Second line')
        for el in d.elements:
            self.assertIsInstance(el, Paragraph)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
test_reader_clean
~~~~~~~~~~~~~~~~~
Test the HTML/XML Cleaner.
"""
import logging
import unittest

from lxml.etree import tostring
from lxml import html

from batterydataextractor.scrape.clean import Cleaner, clean, strip, strip_markup, strip_html


logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


# Input HTML
HTML1 = '''
<html>
  <body>
    <!-- A comment at the start of the document -->
    <div><h1>Heading</h1></div>
    <div><a href="page">Link</a></div>
    <div><div>Nested</div></div>
  </body>
</html>
'''
HTML2 = '<html><head><title>title</title></head><body>Start body<!-- A comment --><script>alert(\'test\');</script><p>This i<span>s</span>s a <strong>test</strong>.Test <a href="link">link</a></body></html>'
HTML3 = '<div><p>First<p>Second<p>Third<p>Fourth<p>Fifth<p>Sixth</div>'
HTML4 = '<p>Here is a text para with some <em>inline</em> markup, like <strong>this</strong>. Also a <a href="#thing">link</a></p>'

# Results from stripping HTML
STRIP1 = '<html>\nHeading\nLink\nNested\n</html>'
STRIP2 = '<html>\ntitle\nStart body\nThis iss a test.Test link\n</html>'
STRIP3 = '<div>\nFirst\nSecond\nThird\nFourth\nFifth\nSixth\n</div>'
STRIP4 = '<p>Here is a text para with some inline markup, like this. Also a link</p>'

# Results from stripping HTML and leaving whitespace to match input
STRIPWS1 = '<html>\n  \n    \n    Heading\n    Link\n    Nested\n  \n</html>'
STRIPWS2 = '<html>titleStart bodyThis iss a test.Test link</html>'
STRIPWS3 = '<div>FirstSecondThirdFourthFifthSixth</div>'
STRIPWS4 = '<p>Here is a text para with some inline markup, like this. Also a link</p>'

# Results of cleaning HTML
CLEAN1 = '<html>\n<body>\n<div>\n<h1>Heading</h1>\n</div>\n<div><a href="page">Link</a></div>\n<div>\n<div>Nested</div>\n</div>\n</body>\n</html>'
CLEAN2 = '<html>\n<head>\n<title>title</title>\n</head>\n<body>Start body\n<p>This i<span>s</span>s a <strong>test</strong>.Test <a href="link">link</a></p>\n</body>\n</html>'
CLEAN3 = '<div>\n<p>First</p>\n<p>Second</p>\n<p>Third</p>\n<p>Fourth</p>\n<p>Fifth</p>\n<p>Sixth</p>\n</div>'
CLEAN4 = '<p>Here is a text para with some <em>inline</em> markup, like <strong>this</strong>. Also a <a href="#thing">link</a></p>'

# Results of stripping HTML and killing strong tags
STRIPKS1 = '<html>\nHeading\nLink\nNested\n</html>'
STRIPKS2 = '<html>\ntitle\nStart body\nThis iss a .Test link\n</html>'
STRIPKS3 = '<div>\nFirst\nSecond\nThird\nFourth\nFifth\nSixth\n</div>'
STRIPKS4 = '<p>Here is a text para with some inline markup, like . Also a link</p>'


class TestCleaner(unittest.TestCase):

    def test_strip1(self):
        """Test applying the default strip to HTML document."""
        tree = html.fromstring(HTML1)
        strip(tree)
        self.assertEqual(STRIP1, tostring(tree).decode())

    def test_strip2(self):
        """Test applying the default strip to HTML document."""
        tree = html.fromstring(HTML2)
        strip(tree)
        self.assertEqual(STRIP2, tostring(tree).decode())

    def test_strip3(self):
        """Test applying the default strip to HTML document."""
        tree = html.fromstring(HTML3)
        strip(tree)
        self.assertEqual(STRIP3, tostring(tree).decode())

    def test_strip4(self):
        """Test applying the default strip to HTML document."""
        tree = html.fromstring(HTML4)
        strip(tree)
        self.assertEqual(STRIP4, tostring(tree).decode())

    def test_strip_markup_doc(self):
        """Test applying the default strip_markup to HTML document."""
        original_tree = html.fromstring(HTML1)
        original_string = tostring(original_tree).decode()
        stripped_tree = strip_markup(original_tree)
        stripped_string = tostring(stripped_tree).decode()
        self.assertEqual(STRIP1, stripped_string)
        self.assertEqual(original_string, tostring(original_tree).decode())  # tree shouldn't change as strip_markup copies

    def test_strip_html1(self):
        """Test applying the default strip_markup to HTML string."""
        self.assertEqual(STRIP1, strip_html(HTML1))

    def test_strip_html2(self):
        """Test applying the default strip_markup to HTML string."""
        self.assertEqual(STRIP2, strip_html(HTML2))

    def test_strip_html3(self):
        """Test applying the default strip_markup to HTML string."""
        self.assertEqual(STRIP3, strip_html(HTML3))

    def test_strip_leave_whitespace1(self):
        """Test applying a ``Cleaner`` instance with ``fix_whitespace=False``."""
        cleaner = Cleaner(fix_whitespace=False, strip_xpath='.//*')
        tree = html.fromstring(HTML1)
        cleaner(tree)
        self.assertEqual(STRIPWS1, tostring(tree).decode())

    def test_strip_leave_whitespace2(self):
        """Test applying a ``Cleaner`` instance with ``fix_whitespace=False``."""
        cleaner = Cleaner(fix_whitespace=False, strip_xpath='.//*')
        tree = html.fromstring(HTML2)
        cleaner(tree)
        self.assertEqual(STRIPWS2, tostring(tree).decode())

    def test_strip_leave_whitespace3(self):
        """Test applying a ``Cleaner`` instance with ``fix_whitespace=False``."""
        cleaner = Cleaner(fix_whitespace=False, strip_xpath='.//*')
        tree = html.fromstring(HTML3)
        cleaner(tree)
        self.assertEqual(STRIPWS3, tostring(tree).decode())

    def test_strip_leave_whitespace4(self):
        """Test applying a ``Cleaner`` instance with ``fix_whitespace=False``."""
        cleaner = Cleaner(fix_whitespace=False, strip_xpath='.//*')
        tree = html.fromstring(HTML4)
        cleaner(tree)
        self.assertEqual(STRIPWS4, tostring(tree).decode())

    def test_clean1(self):
        """Test apply the default clean to HTML document."""
        tree = html.fromstring(HTML1)
        clean(tree)
        self.assertEqual(CLEAN1, tostring(tree).decode())

    def test_clean2(self):
        """Test apply the default clean to HTML document."""
        tree = html.fromstring(HTML2)
        clean(tree)
        self.assertEqual(CLEAN2, tostring(tree).decode())

    def test_clean3(self):
        """Test apply the default clean to HTML document."""
        tree = html.fromstring(HTML3)
        clean(tree)
        self.assertEqual(CLEAN3, tostring(tree, method='html').decode())

    def test_clean4(self):
        """Test apply the default clean to HTML document."""
        tree = html.fromstring(HTML4)
        clean(tree)
        self.assertEqual(CLEAN4, tostring(tree, method='html').decode())

    def test_strip_kill_strong1(self):
        """Test applying a ``Cleaner`` instance that kills strong tags."""
        cleaner = Cleaner(strip_xpath='.//*', kill_xpath='.//script | .//style | .//comment() | .//processing-instruction() | .//strong')
        tree = html.fromstring(HTML1)
        cleaner(tree)
        self.assertEqual(STRIPKS1, tostring(tree).decode())

    def test_strip_kill_strong2(self):
        """Test applying a ``Cleaner`` instance that kills strong tags."""
        cleaner = Cleaner(strip_xpath='.//*', kill_xpath='.//script | .//style | .//comment() | .//processing-instruction() | .//strong')
        tree = html.fromstring(HTML2)
        cleaner(tree)
        self.assertEqual(STRIPKS2, tostring(tree).decode())

    def test_strip_kill_strong3(self):
        """Test applying a ``Cleaner`` instance that kills strong tags."""
        cleaner = Cleaner(strip_xpath='.//*', kill_xpath='.//script | .//style | .//comment() | .//processing-instruction() | .//strong')
        tree = html.fromstring(HTML3)
        cleaner(tree)
        self.assertEqual(STRIPKS3, tostring(tree).decode())

    def test_strip_kill_strong4(self):
        """Test applying a ``Cleaner`` instance that kills strong tags."""
        cleaner = Cleaner(strip_xpath='.//*', kill_xpath='.//script | .//style | .//comment() | .//processing-instruction() | .//strong')
        tree = html.fromstring(HTML4)
        cleaner(tree)
        self.assertEqual(STRIPKS4, tostring(tree).decode())


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.reader.rsc

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
RSC reader
"""
import logging

from ..scrape.clean import clean, Cleaner
from ..scrape.rsc import rsc_html_whitespace, replace_rsc_img_chars, join_rsc_table_captions, rsc_clean_abstract
from .markup import HtmlReader

log = logging.getLogger(__name__)

# XML stripper that removes the tags around numbers in chemical formulas
strip_rsc_html = Cleaner(strip_xpath='.//b')


class RscHtmlReader(HtmlReader):
    """Reader for HTML documents from the RSC."""

    cleaners = [clean, rsc_html_whitespace, replace_rsc_img_chars, rsc_clean_abstract, join_rsc_table_captions, strip_rsc_html]

    root_css = 'html'
    title_css = 'h1, .title_heading'
    heading_css = 'h2, h3, h4, h5, h6, .a_heading, .b_heading, .c_heading, .c_heading_indent, .d_heading,' \
                  '.d_heading_indent'
    abstract_css = 'div[class="abstract"], p[class="abstract"]'
    citation_css = 'span[id^="cit"]'
    reference_css = 'small sup a, a[href^="#cit"], a[href^="#fn"], a[href^="#tab"]'
    ignore_css = '.table_caption + table, .left_head, sup span.sup_ref, small sup a, a[href^="#fn"], .PMedLink, ' \
                 'table, figure'

    def detect(self, fstring, fname=None):
        """"""
        if fname and not (fname.endswith('.html') or fname.endswith('.htm')):
            return False
        if b'meta name="citation_doi" content="10.1039' in fstring:
            return True
        return False



# -*- coding: utf-8 -*-
"""
setup.py
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BatteryDataExtractor setup
author: Shu Huang (sh2009@cam.ac.uk)
"""
from setuptools import setup, find_packages

with open('LICENSE', encoding='utf-8') as f:
    license = f.read()

setup(
    name='batterydataextractor',
    version='0.0.5',
    author='Shu Huang',
    author_email='sh2009@cam.ac.uk',
    license=license,
    url='https://github.com/ShuHuang/batterydataextractor',
    packages=find_packages(),
    description='BatteryDataExtractor: battery-aware text-mining software embedded with BERT models',
    long_description='''BatteryDataExtractor is the first property-specific text-mining tool for auto-generating 
    databases of materials and their property, device, and associated characteristics. The software has been 
    constructed by embedding the BatteryBERT model.''',
    keywords='bde',
    zip_safe=False,
    install_requires=[
        'selenium>=4.1.0',
        'requests>=2.27.1',
        'beautifulsoup4>=4.10.0',
        'numpy>=1.22.0',
        'torch>=1.11.0',
        'torchvision>=0.12.0',
        'torchaudio>=0.11.0',
        'transformers>1.18.0',
        'lxml>=4.7.1',
        'cssselect>=1.1.0',
        'appdirs>=1.4.4',
        'scispacy>=0.4.0',
        'spacy==3.0.7',
        'six'
    ],
    dependency_links=[
    'https://s3-us-west-2.amazonaws.com/ai2-s2-scispacy/releases/v0.4.0/en_core_sci_sm-0.4.0.tar.gz'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Chemistry',
    ],
)



# -*- coding: utf-8 -*-
"""
batterydataextractor.text.processors

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Text processors.
"""
from abc import ABCMeta, abstractmethod
import logging
import six


log = logging.getLogger(__name__)


class BaseProcessor(six.with_metaclass(ABCMeta)):
    """Abstract processor class from which all processors inherit. Subclasses must implement a ``__call__()`` method."""

    @abstractmethod
    def __call__(self, text):
        """Process the text.
        :param string text: The input text.
        :returns: The processed text or None.
        :rtype: string or None
        """
        return text
